import os
import re
import json
import base64
import traceback
import requests
import logging
from six import add_metaclass
from abc import ABCMeta, abstractmethod
from localstack.config import DATA_DIR
from localstack.utils.aws import aws_stack
from localstack.utils.common import to_bytes, to_str
from localstack.utils.bootstrap import is_api_enabled
from localstack.services.generic_proxy import ProxyListener

USE_SINGLE_DUMP_FILE = True

if USE_SINGLE_DUMP_FILE:
    API_FILE_PATTERN = '{data_dir}/recorded_api_calls.json'
else:
    API_FILE_PATTERN = '{data_dir}/{api}_api_calls.json'

# Stack with flags to indicate whether we are currently re-playing API calls.
# (We should not be re-playing and recording at the same time)
CURRENTLY_REPLAYING = []

# file paths by API
API_FILE_PATHS = {}

# flag to indicate if the restoration of api calls is complete
API_CALLS_RESTORED = False

# set up logger
LOG = logging.getLogger(__name__)


@add_metaclass(ABCMeta)
class PersistingProxyListener(ProxyListener):
    """
    This proxy listener could be extended by any API that wishes to record its requests and responses,
    via the existing persistence facility.
    """
    SKIP_PERSISTENCE_TARGET_METHOD_REGEX = re.compile(r'.*\.List|.*\.Describe|.*\.Get')

    def return_response(self, method, path, data, headers, response, request_handler=None):
        res = super(PersistingProxyListener, self).return_response(method, path, data, headers, response,
                                                                   request_handler)

        if self.should_persist(method, path, data, headers, response):
            record(self.api_name(), to_str(method), to_str(path), data, headers, response)

        return res

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def should_persist(self, method, path, data, headers, response):
        """
        Every API listener may choose which endpoints should be persisted;
        The default behavior is persisting all calls with:

        - HTTP PUT / POST / DELETE methods
        - Successful response (non 4xx, 5xx)
        - Excluding methods with 'Describe', 'List', and 'Get' in the X-Amz-Target header

        :param method: The HTTP method name (e.g. 'GET', 'POST')
        :param path: The HTTP path (e.g. '/update')
        :param data: The request body
        :param headers: HTTP response headers
        :param response: HTTP response object
        :return: If True, will persist the current API call.
        :rtype bool
        """
        target_method = headers.get('X-Amz-Target', '')
        skip_target_method = self.SKIP_PERSISTENCE_TARGET_METHOD_REGEX.match(target_method, re.I)

        return should_record(method) and response is not None and response.ok and skip_target_method is None

    @abstractmethod
    def api_name(self):
        """ This should return the name of the API we're operating against, e.g. 'sqs' """
        raise NotImplementedError('Implement me')


def should_record(method):
    """ Decide whether or not a given API call should be recorded (persisted to disk) """
    return method in ['PUT', 'POST', 'DELETE', 'PATCH']


def record(api, method=None, path=None, data=None, headers=None, response=None, request=None):
    """ Record a given API call to a persistent file on disk """
    file_path = get_file_path(api)
    if CURRENTLY_REPLAYING or not file_path:
        return
    if request:
        method = method or request.method
        path = path or request.path
        headers = headers or request.headers
        data = data or request.data

    should_be_recorded = should_record(method)
    if not should_be_recorded:
        return

    try:
        if isinstance(data, dict):
            data = json.dumps(data)

        def get_recordable_data(request_data):
            if request_data or request_data in [u'', b'']:
                try:
                    request_data = to_bytes(request_data)
                except Exception as ex:
                    LOG.warning('Unable to call to_bytes: %s' % ex)
                request_data = to_str(base64.b64encode(request_data))
            return request_data

        data = get_recordable_data(data)
        response_data = get_recordable_data('' if response is None else response.content)

        entry = {
            'a': api,
            'm': method,
            'p': path,
            'd': data,
            'h': dict(headers),
            'rd': response_data
        }

        with open(file_path, 'a') as dumpfile:
            dumpfile.write('%s\n' % json.dumps(entry))
    except Exception as e:
        print('Error recording API call to persistent file: %s %s' % (e, traceback.format_exc()))


def prepare_replay_data(command):
    data = command['d']
    data = data and base64.b64decode(data)
    return data


def replay_command(command):
    api = command['a']
    if not is_api_enabled(api):
        return
    function = getattr(requests, command['m'].lower())
    data = prepare_replay_data(command)
    endpoint = aws_stack.get_local_service_url(api)
    full_url = (endpoint[:-1] if endpoint.endswith('/') else endpoint) + command['p']
    headers = aws_stack.set_internal_auth(command['h'])
    try:
        # fix an error when calling requests with invalid payload encoding
        data and hasattr(data, 'encode') and data.encode('latin-1')
    except UnicodeEncodeError:
        if hasattr(data, 'encode'):
            data = data.encode('utf-8')
    response = function(full_url, data=data, headers=headers, verify=False)
    return response


def replay(api):
    file_path = get_file_path(api)
    if not file_path:
        return
    CURRENTLY_REPLAYING.append(True)
    count = 0
    try:
        with open(file_path, 'r') as reader:
            for line in reader:
                if line.strip():
                    count += 1
                    command = json.loads(line)
                    replay_command(command)
    finally:
        CURRENTLY_REPLAYING.pop(0)
    if count:
        LOG.info('Restored %s API calls from persistent file: %s' % (count, file_path))


def restore_persisted_data(apis):
    global API_CALLS_RESTORED

    if USE_SINGLE_DUMP_FILE:
        replay('_all_')
    else:
        apis = apis if isinstance(apis, list) else [apis]
        for api in apis:
            replay(api)
    API_CALLS_RESTORED = True


# ---------------
# HELPER METHODS
# ---------------

def get_file_path(api, create=True):
    if api not in API_FILE_PATHS:
        API_FILE_PATHS[api] = False
        if not DATA_DIR:
            return False
        file_path = API_FILE_PATTERN.format(data_dir=DATA_DIR, api=api)
        if create and not os.path.exists(file_path):
            with open(file_path, 'a'):
                os.utime(file_path, None)
        if os.path.exists(file_path):
            API_FILE_PATHS[api] = file_path
    return API_FILE_PATHS.get(api)



import os
import re
import json
import time
import logging
from localstack import config
from localstack.constants import TEST_AWS_ACCOUNT_ID, MOTO_ACCOUNT_ID
from localstack.utils.aws import aws_stack
from localstack.utils.common import to_str, save_file, TMP_FILES, mkdir, replace_response_content
from localstack.utils.tagging import TaggingService
from localstack.services.generic_proxy import ProxyListener
from localstack.services.events.scheduler import JobScheduler

LOG = logging.getLogger(__name__)

EVENTS_TMP_DIR = os.path.join(config.TMP_FOLDER, 'cw_events')

# maps rule to job_id
RULE_SCHEDULED_JOBS = {}


def fix_account_id(response):
    return aws_stack.fix_account_id_in_arns(response, existing=MOTO_ACCOUNT_ID, replace=TEST_AWS_ACCOUNT_ID)


def fix_date_format(response):
    """ Normalize date to format '2019-06-13T18:10:09.1234Z' """
    pattern = r'<CreateDate>([^<]+) ([^<+]+)(\+[^<]*)?</CreateDate>'
    replacement = r'<CreateDate>\1T\2Z</CreateDate>'
    replace_response_content(response, pattern, replacement)


def _create_and_register_temp_dir():
    if EVENTS_TMP_DIR not in TMP_FILES:
        mkdir(EVENTS_TMP_DIR)
        TMP_FILES.append(EVENTS_TMP_DIR)


def _dump_events_to_files(events_with_added_uuid):
    current_time_millis = int(round(time.time() * 1000))
    for event in events_with_added_uuid:
        save_file(
            os.path.join(EVENTS_TMP_DIR, '%s_%s' % (current_time_millis, event['uuid'])),
            json.dumps(event['event'])
        )


def get_scheduled_rule_func(data):
    def func(*args):
        rule_name = data.get('Name')
        client = aws_stack.connect_to_service('events')
        targets = client.list_targets_by_rule(Rule=rule_name)['Targets']
        if targets:
            LOG.debug('Notifying %s targets in response to triggered Events rule %s' % (len(targets), rule_name))
        for target in targets:
            arn = target.get('Arn')
            event = json.loads(target.get('Input') or '{}')
            attr = aws_stack.get_events_target_attributes(target)
            aws_stack.send_event_to_target(arn, event, target_attributes=attr)
    return func


def convert_schedule_to_cron(schedule):
    """ Convert Events schedule like "cron(0 20 * * ? *)" or "rate(5 minutes)" """
    cron_regex = r'\s*cron\s*\(([^\)]*)\)\s*'
    if re.match(cron_regex, schedule):
        cron = re.sub(cron_regex, r'\1', schedule)
        return cron
    rate_regex = r'\s*rate\s*\(([^\)]*)\)\s*'
    if re.match(rate_regex, schedule):
        rate = re.sub(rate_regex, r'\1', schedule)
        value, unit = re.split(r'\s+', rate.strip())
        if 'minute' in unit:
            return '*/%s * * * *' % value
        if 'hour' in unit:
            return '* */%s * * *' % value
        if 'day' in unit:
            return '* * */%s * *' % value
        raise Exception('Unable to parse events schedule expression: %s' % schedule)
    return schedule


def handle_put_rule(data):
    schedule = data.get('ScheduleExpression')
    enabled = data.get('State') != 'DISABLED'

    if schedule:
        job_func = get_scheduled_rule_func(data)
        cron = convert_schedule_to_cron(schedule)
        LOG.debug('Adding new scheduled Events rule with cron schedule %s' % cron)

        job_id = JobScheduler.instance().add_job(job_func, cron, enabled)
        region = aws_stack.get_region()
        RULE_SCHEDULED_JOBS[region] = RULE_SCHEDULED_JOBS.get(region) or {}
        RULE_SCHEDULED_JOBS[region][data['Name']] = job_id

    return True


def handle_delete_rule(rule_name):
    region = aws_stack.get_region()
    job_id = RULE_SCHEDULED_JOBS.get(region, {}).get(rule_name)
    if job_id:
        LOG.debug('Removing scheduled Events: {} | job_id: {}'.format(rule_name, job_id))
        JobScheduler.instance().cancel_job(job_id=job_id)


def handle_disable_rule(rule_name):
    region = aws_stack.get_region()
    job_id = RULE_SCHEDULED_JOBS.get(region, {}).get(rule_name)
    if job_id:
        LOG.debug('Disabling Rule: {} | job_id: {}'.format(rule_name, job_id))
        JobScheduler.instance().disable_job(job_id=job_id)


class ProxyListenerEvents(ProxyListener):
    svc = TaggingService()

    def forward_request(self, method, path, data, headers):
        if method == 'OPTIONS':
            return 200

        action = headers.get('X-Amz-Target')
        if method == 'POST' and path == '/':
            parsed_data = json.loads(to_str(data))

            if action == 'AWSEvents.PutRule':
                return handle_put_rule(parsed_data)

            elif action == 'AWSEvents.DeleteRule':
                handle_delete_rule(rule_name=parsed_data.get('Name', None))

            elif action == 'AWSEvents.ListTagsForResource':
                return self.svc.list_tags_for_resource(parsed_data['ResourceARN']) or {}

            elif action == 'AWSEvents.TagResource':
                self.svc.tag_resource(parsed_data['ResourceARN'], parsed_data['Tags'])
                return {}

            elif action == 'AWSEvents.UntagResource':
                self.svc.untag_resource(parsed_data['ResourceARN'], parsed_data['TagKeys'])
                return {}

            elif action == 'AWSEvents.DisableRule':
                handle_disable_rule(rule_name=parsed_data.get('Name', None))

        return True

    def return_response(self, method, path, data, headers, response, request_handler=None):
        if response.content:
            # fix hardcoded account ID in ARNs returned from this API
            fix_account_id(response)

            # fix dates returned from this API (fixes an issue with Terraform)
            fix_date_format(response)

            # fix Content-Length header
            response.headers['Content-Length'] = len(response._content)


# instantiate listener
UPDATE_EVENTS = ProxyListenerEvents()



import re
import json
import time
import logging
import requests
import datetime
from flask import Response as FlaskResponse
from requests.models import Response
from six.moves.urllib_parse import urljoin
from moto.apigateway.models import apigateway_backends
from localstack.utils import common
from localstack.config import TEST_KINESIS_URL, TEST_SQS_URL
from localstack.constants import APPLICATION_JSON, PATH_USER_REQUEST, TEST_AWS_ACCOUNT_ID
from localstack.utils.aws import aws_stack, aws_responses
from localstack.utils.common import to_str, to_bytes
from localstack.utils.analytics import event_publisher
from localstack.services.kinesis import kinesis_listener
from localstack.services.awslambda import lambda_api
from localstack.services.apigateway import helpers
from localstack.services.generic_proxy import ProxyListener
from localstack.utils.aws.aws_responses import flask_to_requests_response, requests_response, LambdaResponse
from localstack.services.apigateway.helpers import (get_resource_for_path, handle_authorizers, handle_validators,
    handle_accounts, extract_query_string_params, extract_path_params, make_error_response, get_cors_response,
    hande_base_path_mappings)

# set up logger
LOGGER = logging.getLogger(__name__)

# regex path patterns
PATH_REGEX_AUTHORIZERS = r'^/restapis/([A-Za-z0-9_\-]+)/authorizers(\?.*)?'
PATH_REGEX_VALIDATORS = r'^/restapis/([A-Za-z0-9_\-]+)/requestvalidators(\?.*)?'
PATH_REGEX_RESPONSES = r'^/restapis/([A-Za-z0-9_\-]+)/gatewayresponses(/[A-Za-z0-9_\-]+)?(\?.*)?'
PATH_REGEX_USER_REQUEST = r'^/restapis/([A-Za-z0-9_\-]+)/([A-Za-z0-9_\-]+)/%s/(.*)$' % PATH_USER_REQUEST
PATH_REGEX_PATH_MAPPINGS = r'/domainnames/([^/]+)/basepathmappings(/.*)?'
HOST_REGEX_EXECUTE_API = r'(.*://)?([a-zA-Z0-9-]+)\.execute-api\..*'

# Maps API IDs to list of gateway responses
GATEWAY_RESPONSES = {}


class AuthorizationError(Exception):
    pass


class ProxyListenerApiGateway(ProxyListener):
    def forward_request(self, method, path, data, headers):
        if re.match(PATH_REGEX_USER_REQUEST, path):
            return invoke_rest_api_from_request(method, path, data, headers)

        data = data and json.loads(to_str(data))

        if re.match(PATH_REGEX_AUTHORIZERS, path):
            return handle_authorizers(method, path, data, headers)

        if re.match(PATH_REGEX_VALIDATORS, path):
            return handle_validators(method, path, data, headers)

        if re.match(PATH_REGEX_RESPONSES, path):
            search_match = re.search(PATH_REGEX_RESPONSES, path)
            api_id = search_match.group(1)
            response_type = (search_match.group(2) or '').lstrip('/')
            if method == 'GET':
                if response_type:
                    return get_gateway_response(api_id, response_type)
                return get_gateway_responses(api_id)
            if method == 'PUT':
                return put_gateway_response(api_id, response_type, data)

        return True

    def return_response(self, method, path, data, headers, response):
        # fix backend issue (missing support for API documentation)
        if re.match(r'/restapis/[^/]+/documentation/versions', path):
            if response.status_code == 404:
                return requests_response({'position': '1', 'items': []})

        # add missing implementations
        if response.status_code == 404:
            data = data and json.loads(to_str(data))
            result = None
            if path == '/account':
                result = handle_accounts(method, path, data, headers)
            if re.match(PATH_REGEX_PATH_MAPPINGS, path):
                result = hande_base_path_mappings(method, path, data, headers)
            if result is not None:
                response.status_code = 200
                aws_responses.set_response_content(response, result)

        # publish event
        if method == 'POST' and path == '/restapis':
            content = json.loads(to_str(response.content))
            event_publisher.fire_event(event_publisher.EVENT_APIGW_CREATE_API,
                payload={'a': event_publisher.get_hash(content['id'])})
        api_regex = r'^/restapis/([a-zA-Z0-9\-]+)$'
        if method == 'DELETE' and re.match(api_regex, path):
            api_id = re.sub(api_regex, r'\1', path)
            event_publisher.fire_event(event_publisher.EVENT_APIGW_DELETE_API,
                payload={'a': event_publisher.get_hash(api_id)})


# ------------
# API METHODS
# ------------

def get_gateway_responses(api_id):
    result = GATEWAY_RESPONSES.get(api_id, [])
    base_path = '/restapis/%s/gatewayresponses' % api_id
    href = 'http://docs.aws.amazon.com/apigateway/latest/developerguide/restapi-gatewayresponse-{rel}.html'

    def item(i):
        i['_links'] = {
            'self': {
                'href': '%s/%s' % (base_path, i['responseType'])
            },
            'gatewayresponse:put': {
                'href': '%s/{response_type}' % base_path,
                'templated': True
            },
            'gatewayresponse:update': {
                'href': '%s/%s' % (base_path, i['responseType'])
            }
        }
        i['responseParameters'] = i.get('responseParameters', {})
        i['responseTemplates'] = i.get('responseTemplates', {})
        return i

    result = {
        '_links': {
            'curies': {
                'href': href,
                'name': 'gatewayresponse',
                'templated': True
            },
            'self': {'href': base_path},
            'first': {'href': base_path},
            'gatewayresponse:by-type': {
                'href': '%s/{response_type}' % base_path,
                'templated': True
            },
            'item': [{'href': '%s/%s' % (base_path, r['responseType'])} for r in result]
        },
        '_embedded': {
            'item': [item(i) for i in result]
        },
        # Note: Looks like the format required by aws CLI ("item" at top level) differs from the docs:
        # https://docs.aws.amazon.com/apigateway/api-reference/resource/gateway-responses/
        'item': [item(i) for i in result]
    }
    return result


def get_gateway_response(api_id, response_type):
    responses = GATEWAY_RESPONSES.get(api_id, [])
    result = [r for r in responses if r['responseType'] == response_type]
    return result[0] if result else 404


def put_gateway_response(api_id, response_type, data):
    GATEWAY_RESPONSES[api_id] = GATEWAY_RESPONSES.get(api_id, [])
    data['responseType'] = response_type
    GATEWAY_RESPONSES[api_id].append(data)
    return data


def run_authorizer(api_id, headers, authorizer):
    # TODO implement authorizers
    pass


def authorize_invocation(api_id, headers):
    client = aws_stack.connect_to_service('apigateway')
    authorizers = client.get_authorizers(restApiId=api_id, limit=100).get('items', [])
    for authorizer in authorizers:
        run_authorizer(api_id, headers, authorizer)


def validate_api_key(api_key, stage):

    usage_plan_ids = []

    client = aws_stack.connect_to_service('apigateway')
    usage_plans = client.get_usage_plans()
    for item in usage_plans.get('items', []):
        api_stages = item.get('apiStages', [])
        for api_stage in api_stages:
            if api_stage.get('stage') == stage:
                usage_plan_ids.append(item.get('id'))

    for usage_plan_id in usage_plan_ids:
        usage_plan_keys = client.get_usage_plan_keys(usagePlanId=usage_plan_id)
        for key in usage_plan_keys.get('items', []):
            if key.get('value') == api_key:
                return True

    return False


def is_api_key_valid(is_api_key_required, headers, stage):
    if not is_api_key_required:
        return True

    api_key = headers.get('X-API-Key')
    if not api_key:
        return False

    return validate_api_key(api_key, stage)


def update_content_length(response):
    if response and response.content:
        response.headers['Content-Length'] = str(len(response.content))


def apply_template(integration, req_res_type, data, path_params={}, query_params={}, headers={}):
    integration_type = integration.get('type') or integration.get('integrationType')
    if integration_type in ['HTTP', 'AWS']:
        # apply custom request template
        template = integration.get('%sTemplates' % req_res_type, {}).get(APPLICATION_JSON)
        if template:
            context = {}
            context['body'] = data

            def _params(name=None):
                # See https://docs.aws.amazon.com/apigateway/latest/developerguide/
                #    api-gateway-mapping-template-reference.html#input-variable-reference
                # Returns "request parameter from the path, query string, or header value (searched in that order)"
                combined = {}
                combined.update(path_params or {})
                combined.update(query_params or {})
                combined.update(headers or {})
                return combined if not name else combined.get(name)

            context['params'] = _params
            data = aws_stack.render_velocity_template(template, context)
    return data


def get_api_id_stage_invocation_path(path, headers):
    path_match = re.search(PATH_REGEX_USER_REQUEST, path)
    host_header = headers.get('Host', '')
    host_match = re.search(HOST_REGEX_EXECUTE_API, host_header)
    if path_match:
        api_id = path_match.group(1)
        stage = path_match.group(2)
        relative_path_w_query_params = '/%s' % path_match.group(3)
    elif host_match:
        api_id = host_match.group(1)
        stage = path.strip('/').split('/')[0]
        relative_path_w_query_params = '/%s' % path.lstrip('/').partition('/')[2]
    return api_id, stage, relative_path_w_query_params


def invoke_rest_api_from_request(method, path, data, headers, context={}):
    api_id, stage, relative_path_w_query_params = get_api_id_stage_invocation_path(path, headers)
    try:
        return invoke_rest_api(api_id, stage, method, relative_path_w_query_params,
            data, headers, path=path, context=context)
    except AuthorizationError as e:
        return make_error_response('Not authorized to invoke REST API %s: %s' % (api_id, e), 403)


def invoke_rest_api(api_id, stage, method, invocation_path, data, headers, path=None, context={}):
    path = path or invocation_path
    relative_path, query_string_params = extract_query_string_params(path=invocation_path)

    # run gateway authorizers for this request
    authorize_invocation(api_id, headers)
    path_map = helpers.get_rest_api_paths(rest_api_id=api_id)
    try:
        extracted_path, resource = get_resource_for_path(path=relative_path, path_map=path_map)
    except Exception:
        return make_error_response('Unable to find path %s' % path, 404)

    api_key_required = resource.get('resourceMethods', {}).get(method, {}).get('apiKeyRequired')
    if not is_api_key_valid(api_key_required, headers, stage):
        return make_error_response('Access denied - invalid API key', 403)

    integrations = resource.get('resourceMethods', {})
    integration = integrations.get(method, {})
    if not integration:
        integration = integrations.get('ANY', {})
    integration = integration.get('methodIntegration')
    if not integration:
        if method == 'OPTIONS' and 'Origin' in headers:
            # default to returning CORS headers if this is an OPTIONS request
            return get_cors_response(headers)
        return make_error_response('Unable to find integration for path %s' % path, 404)

    res_methods = path_map.get(relative_path, {}).get('resourceMethods', {})
    meth_integration = res_methods.get(method, {}).get('methodIntegration', {})
    int_responses = meth_integration.get('integrationResponses', {})
    response_templates = int_responses.get('200', {}).get('responseTemplates', {})

    return invoke_rest_api_integration(api_id, stage, integration, method, path, invocation_path, data,
        headers, resource_path=extracted_path, context=context,
        resource_id=resource.get('id'), response_templates=response_templates)


def invoke_rest_api_integration(api_id, stage, integration, method, path, invocation_path, data, headers,
        resource_path, context={}, resource_id=None, response_templates={}):

    relative_path, query_string_params = extract_query_string_params(path=invocation_path)
    integration_type_orig = integration.get('type') or integration.get('integrationType') or ''
    integration_type = integration_type_orig.upper()
    uri = integration.get('uri') or integration.get('integrationUri')

    if (uri.startswith('arn:aws:apigateway:') and ':lambda:path' in uri) or uri.startswith('arn:aws:lambda'):
        if integration_type in ['AWS', 'AWS_PROXY']:
            func_arn = uri
            if ':lambda:path' in uri:
                func_arn = uri.split(':lambda:path')[1].split('functions/')[1].split('/invocations')[0]
            data_str = json.dumps(data) if isinstance(data, (dict, list)) else to_str(data)

            try:
                path_params = extract_path_params(path=relative_path, extracted_path=resource_path)
            except Exception:
                path_params = {}

            # apply custom request template
            data_str = apply_template(integration, 'request', data_str, path_params=path_params,
                query_params=query_string_params, headers=headers)

            # Sample request context:
            # https://docs.aws.amazon.com/apigateway/latest/developerguide/api-gateway-create-api-as-simple-proxy-for-lambda.html#api-gateway-create-api-as-simple-proxy-for-lambda-test
            request_context = get_lambda_event_request_context(method, path, data, headers,
                integration_uri=uri, resource_id=resource_id)
            stage_variables = get_stage_variables(api_id, stage)

            result = lambda_api.process_apigateway_invocation(func_arn, relative_path, data_str,
                stage, api_id, headers, path_params=path_params, query_string_params=query_string_params,
                method=method, resource_path=resource_path, request_context=request_context,
                event_context=context, stage_variables=stage_variables)

            if isinstance(result, FlaskResponse):
                response = flask_to_requests_response(result)
            elif isinstance(result, Response):
                response = result
            else:
                response = LambdaResponse()
                parsed_result = result if isinstance(result, dict) else json.loads(str(result or '{}'))
                parsed_result = common.json_safe(parsed_result)
                parsed_result = {} if parsed_result is None else parsed_result
                response.status_code = int(parsed_result.get('statusCode', 200))
                parsed_headers = parsed_result.get('headers', {})
                if parsed_headers is not None:
                    response.headers.update(parsed_headers)
                try:
                    if isinstance(parsed_result['body'], dict):
                        response._content = json.dumps(parsed_result['body'])
                    else:
                        response._content = to_bytes(parsed_result['body'])
                except Exception:
                    response._content = '{}'
                update_content_length(response)
                response.multi_value_headers = parsed_result.get('multiValueHeaders') or {}

            # apply custom response template
            response._content = apply_template(integration, 'response', response._content)
            response.headers['Content-Length'] = str(len(response.content or ''))

            return response

        msg = 'API Gateway %s integration action "%s", method "%s" not yet implemented' % (
            integration_type, uri, method)
        LOGGER.warning(msg)
        return make_error_response(msg, 404)

    elif integration_type == 'AWS':
        if 'kinesis:action/' in uri:
            if uri.endswith('kinesis:action/PutRecords'):
                target = kinesis_listener.ACTION_PUT_RECORDS
            if uri.endswith('kinesis:action/ListStreams'):
                target = kinesis_listener.ACTION_LIST_STREAMS

            template = integration['requestTemplates'][APPLICATION_JSON]
            new_request = aws_stack.render_velocity_template(template, data)
            # forward records to target kinesis stream
            headers = aws_stack.mock_aws_request_headers(service='kinesis')
            headers['X-Amz-Target'] = target
            result = common.make_http_request(url=TEST_KINESIS_URL,
                method='POST', data=new_request, headers=headers)
            # TODO apply response template..?
            return result

        elif 'states:action/' in uri:
            if uri.endswith('states:action/StartExecution'):
                action = 'StartExecution'
            decoded_data = data.decode()
            if 'stateMachineArn' in decoded_data and 'input' in decoded_data:
                payload = json.loads(decoded_data)
            elif APPLICATION_JSON in integration.get('requestTemplates', {}):
                template = integration['requestTemplates'][APPLICATION_JSON]
                payload = aws_stack.render_velocity_template(template, data, as_json=True)
            client = aws_stack.connect_to_service('stepfunctions')

            kwargs = {'name': payload['name']} if 'name' in payload else {}
            result = client.start_execution(stateMachineArn=payload['stateMachineArn'],
                            input=payload['input'], **kwargs)
            response = requests_response(
                content={
                    'executionArn': result['executionArn'],
                    'startDate': str(result['startDate'])
                },
                headers=aws_stack.mock_aws_request_headers()
            )
            return response

        if method == 'POST':
            if uri.startswith('arn:aws:apigateway:') and ':sqs:path' in uri:
                template = integration['requestTemplates'][APPLICATION_JSON]
                account_id, queue = uri.split('/')[-2:]
                region_name = uri.split(':')[3]

                new_request = '%s&QueueName=%s' % (aws_stack.render_velocity_template(template, data), queue)
                headers = aws_stack.mock_aws_request_headers(service='sqs', region_name=region_name)

                url = urljoin(TEST_SQS_URL, '%s/%s' % (TEST_AWS_ACCOUNT_ID, queue))
                result = common.make_http_request(url, method='POST', headers=headers, data=new_request)
                return result

        msg = 'API Gateway AWS integration action URI "%s", method "%s" not yet implemented' % (uri, method)
        LOGGER.warning(msg)
        return make_error_response(msg, 404)

    elif integration_type == 'AWS_PROXY':
        if uri.startswith('arn:aws:apigateway:') and ':dynamodb:action' in uri:
            # arn:aws:apigateway:us-east-1:dynamodb:action/PutItem&Table=MusicCollection
            table_name = uri.split(':dynamodb:action')[1].split('&Table=')[1]
            action = uri.split(':dynamodb:action')[1].split('&Table=')[0]

            if 'PutItem' in action and method == 'PUT':
                response_template = response_templates.get('application/json', None)

                if response_template is None:
                    msg = 'Invalid response template defined in integration response.'
                    return make_error_response(msg, 404)

                response_template = json.loads(response_template)
                if response_template['TableName'] != table_name:
                    msg = 'Invalid table name specified in integration response template.'
                    return make_error_response(msg, 404)

                dynamo_client = aws_stack.connect_to_resource('dynamodb')
                table = dynamo_client.Table(table_name)

                event_data = {}
                data_dict = json.loads(data)
                for key, _ in response_template['Item'].items():
                    event_data[key] = data_dict[key]

                table.put_item(Item=event_data)
                response = requests_response(event_data, headers=aws_stack.mock_aws_request_headers())
                return response
        else:
            msg = 'API Gateway action uri "%s", integration type %s not yet implemented' % (uri, integration_type)
            LOGGER.warning(msg)
            return make_error_response(msg, 404)

    elif integration_type in ['HTTP_PROXY', 'HTTP']:
        function = getattr(requests, method.lower())

        # apply custom request template
        data = apply_template(integration, 'request', data)

        if isinstance(data, dict):
            data = json.dumps(data)

        result = function(uri, data=data, headers=headers)

        # apply custom response template
        data = apply_template(integration, 'response', data)

        return result

    elif integration_type == 'MOCK':
        # TODO: add logic for MOCK responses
        pass

    if method == 'OPTIONS':
        # fall back to returning CORS headers if this is an OPTIONS request
        return get_cors_response(headers)

    msg = ('API Gateway integration type "%s", method "%s", URI "%s" not yet implemented' %
           (integration_type, method, uri))
    LOGGER.warning(msg)
    return make_error_response(msg, 404)


def get_stage_variables(api_id, stage):
    region_name = [name for name, region in apigateway_backends.items() if api_id in region.apis][0]
    api_gateway_client = aws_stack.connect_to_service('apigateway', region_name=region_name)
    response = api_gateway_client.get_stage(restApiId=api_id, stageName=stage)
    return response.get('variables', None)


def get_lambda_event_request_context(method, path, data, headers, integration_uri=None, resource_id=None):
    _, stage, relative_path_w_query_params = get_api_id_stage_invocation_path(path, headers)
    relative_path, query_string_params = extract_query_string_params(path=relative_path_w_query_params)
    source_ip = headers.get('X-Forwarded-For', ',').split(',')[-2].strip()
    integration_uri = integration_uri or ''
    account_id = integration_uri.split(':lambda:path')[-1].split(':function:')[0].split(':')[-1]
    request_context = {
        # adding stage to the request context path.
        # https://github.com/localstack/localstack/issues/2210
        'path': '/' + stage + relative_path,
        'accountId': account_id,
        'resourceId': resource_id,
        'stage': stage,
        'identity': {
            'accountId': account_id,
            'sourceIp': source_ip,
            'userAgent': headers.get('User-Agent'),
        },
        'httpMethod': method,
        'protocol': 'HTTP/1.1',
        'requestTime': datetime.datetime.utcnow(),
        'requestTimeEpoch': int(time.time() * 1000),
    }
    return request_context


# instantiate listener
UPDATE_APIGATEWAY = ProxyListenerApiGateway()



import pstats
import logging
import cProfile
import threading
import traceback
from functools import wraps
from localstack.constants import LOCALSTACK_ROOT_FOLDER

# set up logger
LOG = logging.getLogger(__name__)

PROFILED_METHODS = [
    # TODO activate as needed
    # (str, 'lower'),
    # (str, 'rpartition', False),
    # (_io.BufferedReader, 'readline', True)
]


def profiled(lines=50):
    """ Function decorator that profiles code execution. """
    # return profiled_via_yappi(lines)
    return profiled_via_cprofile(lines)


def profiled_via_cprofile(lines=50):
    skipped_lines = ['site-packages', 'lib/python']
    Thread = threading.Thread
    method_invocations = {}

    for m_entry in PROFILED_METHODS:

        def _patch(entry):
            m_obj = getattr(entry[0], entry[1])
            m_str = str(m_obj)
            method_invocations[m_str] = {'__count__': 0}

            def m_patched(*args, **kwargs):
                entry = method_invocations[m_str]
                stack = traceback.format_stack()
                last_frame = str(stack[-2]).strip()
                entry[last_frame] = entry.get(last_frame, 0) + 1
                if entry['__count__'] % 1000 == 0:
                    # TODO print(''.join(stack))
                    pass
                entry['__count__'] += 1
                return m_obj(*args, **kwargs)

            try:
                from forbiddenfruit import curse
                curse(entry[0], entry[1], m_patched)
            except Exception as e:
                print('Unable to set attr:', entry[0], entry[1], m_patched, e)

        _patch(m_entry)

    def add_stats(prof):
        if Thread.stats is None:
            Thread.stats = pstats.Stats(prof)
        else:
            Thread.stats.add(prof)

    def enable_thread_profiling():
        if getattr(Thread, '_profiling_patched', None):
            return
        setattr(Thread, '_profiling_patched', True)
        Thread.stats = None
        thread_run = Thread.run

        def profile_run(self):
            self._prof = cProfile.Profile()
            self._prof.enable()
            thread_run(self)
            self._prof.disable()
            add_stats(self._prof)

        Thread.run = profile_run

    def profile_current_thread(close_profile=None):
        if close_profile:
            close_profile.disable()
            add_stats(close_profile)
            return
        prof = cProfile.Profile()
        prof.enable()
        return prof

    def get_thread_stats():
        stats = getattr(Thread, 'stats', None)
        if stats is None:
            LOG.warning('Thread profiling was not enabled, or no threads finished running.')
        return stats

    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            enable_thread_profiling()
            prof_this_thread = profile_current_thread()
            try:
                return f(*args, **kwargs)
            finally:
                profile_current_thread(prof_this_thread)
                result = get_thread_stats()
                list_orig = result.fcn_list
                for sort in ('tottime', 'cumulative', 'ncalls'):
                    result.fcn_list = list_orig
                    result.sort_stats(sort)
                    result.fcn_list = [e for e in result.fcn_list
                        if not any([s in str(e) for s in skipped_lines])]
                    result.print_stats(lines)
        return wrapped
    return wrapper


def profiled_via_yappi(lines=50):
    skipped_lines = ['site-packages', 'lib/python']
    skipped_lines = []

    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            import yappi
            yappi.start()
            try:
                return f(*args, **kwargs)
            finally:
                result = list(yappi.get_func_stats())
                yappi.stop()
                yappi.clear_stats()
                result = [r for r in result if all([s not in r.full_name for s in skipped_lines])]
                entries = result[:lines]
                prefix = LOCALSTACK_ROOT_FOLDER
                result = []
                result.append('ncall\tttot\ttsub\ttavg\tname')

                def c(num):
                    return str(num)[:7]

                for e in entries:
                    name = e.full_name.replace(prefix, '')
                    result.append('%s\t%s\t%s\t%s\t%s' % (c(e.ncall), c(e.ttot), c(e.tsub), c(e.tavg), name))
                result = '\n'.join(result)
                print(result)
        return wrapped
    return wrapper


def log_duration(name=None):
    """ Function decorator to log the duration of function invocations. """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            from localstack.utils.common import now_utc
            start_time = now_utc(millis=True)
            try:
                return f(*args, **kwargs)
            finally:
                end_time = now_utc(millis=True)
                func_name = name or f.__name__
                duration = (end_time - start_time) * 1000
                if duration > 500:
                    LOG.info('Execution of "%s" took %sms' % (func_name, duration))
        return wrapped
    return wrapper



import re
from localstack.utils.common import to_str, to_bytes
from localstack.services.generic_proxy import ProxyListener


class ProxyListenerSES(ProxyListener):

    def return_response(self, method, path, data, headers, response):
        xml = to_str(response.content)
        has_xmlns = self.response_has_xmlns(xml)
        if xml and xml[0] == '<' and not has_xmlns:
            xmlns = 'http://ses.amazonaws.com/doc/2010-01-31/'
            response_type = re.findall(r'<(\w+)\s?>', xml)[0]
            response_xmlns = '%s xmlns="%s"' % (response_type, xmlns)

            new_xml = re.sub(r'<\?xml.+\?>\n\s*', '', xml)
            new_xml = re.sub(r'</?Errors>\n\s*', '', new_xml)
            new_xml = re.sub(response_type, response_xmlns, new_xml, 1)

            response._content = to_bytes(new_xml)
            response.headers.update({'Content-Length': str(len(new_xml))})
        return response

    def response_has_xmlns(self, xml_string):
        return re.search(r'\s+xmlns="\S+"', xml_string)


# instantiate listener
UPDATE_SES = ProxyListenerSES()



import os
import json
from flask import Flask, render_template, jsonify, send_from_directory, request
from flask_swagger import swagger
from localstack import config
from localstack.utils import common
from localstack.services import generic_proxy, plugins
from localstack.services import infra as services_infra
from localstack.constants import VERSION, LOCALSTACK_WEB_PROCESS
from localstack.dashboard import infra
from localstack.utils.bootstrap import load_plugins, canonicalize_api_names
from localstack.utils.aws.aws_stack import Environment


root_path = os.path.dirname(os.path.realpath(__file__))
web_dir = root_path + '/web/'

app = Flask('app', template_folder=web_dir)
app.root_path = root_path


@app.route('/swagger.json')
def spec():
    swag = swagger(app)
    swag['info']['version'] = VERSION
    swag['info']['title'] = 'AWS Resources Dashboard'
    return jsonify(swag)


@app.route('/graph', methods=['POST'])
def get_graph():
    # TODO remove?
    """ Get deployment graph
        ---
        operationId: 'getGraph'
        parameters:
            - name: request
              in: body
    """
    data = get_payload()
    env = Environment.from_string(data.get('awsEnvironment'))
    graph = infra.get_graph(name_filter=data['nameFilter'], env=env, region=data.get('awsRegion'))
    return jsonify(graph)


@app.route('/services', methods=['GET'])
def get_status():
    """ Get status of deployed services
        ---
        operationId: 'getStatus'
    """
    result = services_infra.get_services_status()
    return jsonify(result)


@app.route('/services', methods=['POST'])
def set_status():
    """ Set status of deployed services
        ---
        operationId: 'setStatus'
    """
    data = get_payload()
    result = services_infra.set_service_status(data)
    return jsonify(result)


@app.route('/kinesis/<streamName>/<shardId>/events/latest', methods=['POST'])
def get_kinesis_events(streamName, shardId):
    """ Get latest events from Kinesis.
        ---
        operationId: 'getKinesisEvents'
        parameters:
            - name: streamName
              in: path
            - name: shardId
              in: path
            - name: request
              in: body
    """
    data = get_payload()
    env = Environment.from_string(data.get('awsEnvironment'))
    result = infra.get_kinesis_events(stream_name=streamName, shard_id=shardId, env=env)
    return jsonify(result)


@app.route('/lambda/<functionName>/code', methods=['POST'])
def get_lambda_code(functionName):
    """ Get source code for Lambda function.
        ---
        operationId: 'getLambdaCode'
        parameters:
            - name: functionName
              in: path
            - name: request
              in: body
    """
    data = get_payload()
    env = Environment.from_string(data.get('awsEnvironment'))
    result = infra.get_lambda_code(func_name=functionName, env=env)
    return jsonify(result)


@app.route('/health')
def get_health():
    # TODO: this should be moved into a separate service, once the dashboard UI is removed
    reload = request.args.get('reload') is not None
    result = plugins.get_services_health(reload=reload)
    return jsonify(result)


@app.route('/')
def hello():
    return render_template('index.html')


@app.route('/<path:path>')
def send_static(path):
    return send_from_directory(web_dir + '/', path)


def get_payload():
    return json.loads(common.to_str(request.data))


def ensure_webapp_installed():
    web_dir = os.path.realpath(os.path.join(os.path.dirname(__file__), 'web'))
    node_modules_dir = os.path.join(web_dir, 'node_modules', 'jquery')
    if not os.path.exists(node_modules_dir):
        print('Initializing installation of Web application (this could take a long time, please be patient)')
        common.run('cd "%s"; npm install' % web_dir)


def serve(port):
    os.environ[LOCALSTACK_WEB_PROCESS] = '1'
    ensure_webapp_installed()
    load_plugins()
    canonicalize_api_names()

    backend_url = 'http://localhost:%s' % port
    services_infra.start_proxy(config.PORT_WEB_UI_SSL, backend_url, use_ssl=True)

    generic_proxy.serve_flask_app(app=app, port=port, quiet=True)



from localstack import config
from moto.route53 import responses as route53_responses
from moto.route53.models import route53_backend
from six.moves.urllib.parse import parse_qs, urlparse
from localstack.services.infra import start_moto_server


def apply_patches():
    # patch function to match hosted zone if zone name ends with a dot
    def list_hosted_zones_by_name_response(self, request, full_url, headers):
        parsed_url = urlparse(full_url)
        query_params = parse_qs(parsed_url.query)
        dnsname = query_params.get('dnsname')
        dnsname = dnsname and dnsname[0]
        all_zones = route53_backend.get_all_hosted_zones()
        zones1 = [zone for zone in all_zones if zone.name == dnsname]
        zones2 = [zone for zone in all_zones if zone.name == '%s.' % dnsname]
        if not zones1 and zones2:
            full_url = full_url.replace('dnsname=%s' % dnsname, 'dnsname=%s.' % dnsname)
        return list_hosted_zones_by_name_response_orig(self, request, full_url, headers)

    list_hosted_zones_by_name_response_orig = route53_responses.Route53.list_hosted_zones_by_name_response
    route53_responses.Route53.list_hosted_zones_by_name_response = list_hosted_zones_by_name_response


def start_route53(port=None, asynchronous=False, update_listener=None):
    port = port or config.PORT_ROUTE53
    apply_patches()
    return start_moto_server(
        'route53', port, name='Route53',
        asynchronous=asynchronous, update_listener=update_listener
    )



import logging
import os
import traceback
from localstack import config
from localstack.services import install
from localstack.utils.aws import aws_stack
from localstack.utils.common import mkdir, wait_for_port_open, get_free_tcp_port
from localstack.services.infra import start_proxy_for_service, do_run, log_startup_message
from localstack.services.install import ROOT_PATH

LOGGER = logging.getLogger(__name__)

# backend service port (updated on startup)
PORT_DYNAMODB_BACKEND = None


def check_dynamodb(expect_shutdown=False, print_error=False):
    out = None
    try:
        # wait for backend port to be opened
        wait_for_port_open(PORT_DYNAMODB_BACKEND, http_path='/', expect_success=False, sleep_time=1)
        # check DynamoDB
        out = aws_stack.connect_to_service('dynamodb').list_tables()
    except Exception as e:
        if print_error:
            LOGGER.error('DynamoDB health check failed: %s %s' % (e, traceback.format_exc()))
    if expect_shutdown:
        assert out is None
    else:
        assert isinstance(out['TableNames'], list)


def start_dynamodb(port=None, asynchronous=False, update_listener=None):
    global PORT_DYNAMODB_BACKEND
    PORT_DYNAMODB_BACKEND = get_free_tcp_port()
    port = port or config.PORT_DYNAMODB
    install.install_dynamodb_local()
    ddb_data_dir_param = '-inMemory'
    if config.DATA_DIR:
        ddb_data_dir = '%s/dynamodb' % config.DATA_DIR
        mkdir(ddb_data_dir)
        # as the service command cds into a different directory, the absolute
        # path of the DATA_DIR is needed as the -dbPath
        absolute_path = os.path.abspath(ddb_data_dir)
        ddb_data_dir_param = '-dbPath %s' % absolute_path
    cmd = ('cd %s/infra/dynamodb/; java -Djava.library.path=./DynamoDBLocal_lib ' +
        '-Xmx%s -jar DynamoDBLocal.jar -port %s %s') % (
        ROOT_PATH, config.DYNAMODB_HEAP_SIZE, PORT_DYNAMODB_BACKEND, ddb_data_dir_param)
    log_startup_message('DynamoDB')
    start_proxy_for_service('dynamodb', port, backend_port=PORT_DYNAMODB_BACKEND, update_listener=update_listener)
    return do_run(cmd, asynchronous, auto_restart=True)



# NOTE: This code is deprecated and will be removed in a future iteration!

import sys
import json
import types
import logging
import traceback
import six
import boto3.session
from moto.s3 import models as s3_models
from moto.iam import models as iam_models
from moto.sqs import models as sqs_models
from moto.sns import models as sns_models
from moto.core import BaseModel
from moto.server import main as moto_main
from moto.kinesis import models as kinesis_models
from moto.dynamodb import models as dynamodb_models
from moto.dynamodb2 import models as dynamodb2_models
from moto.awslambda import models as lambda_models
from moto.apigateway import models as apigw_models
from moto.cloudwatch import models as cw_models
from moto.cloudformation import parsing, responses
from moto.cloudformation import utils as cloudformation_utils
from moto.cloudformation import models as cloudformation_models
from boto.cloudformation.stack import Output
from moto.cloudformation.models import FakeStack, FakeChangeSet, CloudFormationBackend, cloudformation_backends
from moto.cloudformation.exceptions import ValidationError, UnformattedGetAttTemplateException
from localstack import config
from localstack.constants import TEST_AWS_ACCOUNT_ID, MOTO_ACCOUNT_ID
from localstack.utils.aws import aws_stack, aws_responses
from localstack.utils.common import (
    FuncThread, short_uid, recurse_object, clone, json_safe, get_free_tcp_port, Mock, start_thread, clone_safe)
from localstack.services.infra import (
    start_proxy_for_service, do_run, canonicalize_api_names, log_startup_message)
from localstack.utils.bootstrap import setup_logging
from localstack.utils.cloudformation import template_deployer_old as template_deployer
from localstack.services.cloudformation import service_models
from localstack.utils.cloudformation.template_deployer import (
    DependencyNotYetSatisfied, add_default_resource_props)

LOG = logging.getLogger(__name__)

MOTO_CFN_ACCOUNT_ID = '123456789'

# Maps (stack_name,resource_logical_id) -> Bool to indicate which resources are currently being updated
CURRENTLY_UPDATING_RESOURCES = {}

# whether to start the API in a separate process
RUN_SERVER_IN_PROCESS = False

# maximum depth of the resource dependency tree
MAX_DEPENDENCY_DEPTH = 40

# map of additional model classes
MODEL_MAP = {
    'AWS::StepFunctions::Activity': service_models.SFNActivity,
    'AWS::SNS::Subscription': service_models.SNSSubscription,
    'AWS::ApiGateway::GatewayResponse': service_models.GatewayResponse,
    'AWS::ApiGateway::Deployment': apigw_models.Deployment,
    'AWS::ApiGateway::Method': apigw_models.Method,
    'AWS::ApiGateway::Resource': apigw_models.Resource,
    'AWS::ApiGateway::RestApi': apigw_models.RestAPI,
    'AWS::ApiGateway::Stage': apigw_models.Stage,
    'AWS::CloudFormation::Stack': service_models.CloudFormationStack,
    'AWS::SSM::Parameter': service_models.SSMParameter,
    'AWS::Logs::LogGroup': service_models.LogsLogGroup,
    'AWS::KinesisFirehose::DeliveryStream': service_models.FirehoseDeliveryStream,
    'AWS::SecretsManager::Secret': service_models.SecretsManagerSecret,
    'AWS::Elasticsearch::Domain': service_models.ElasticsearchDomain,
    'AWS::Events::Rule': service_models.EventsRule,
    'AWS::S3::BucketPolicy': service_models.S3BucketPolicy
}


def start_cloudformation(port=None, asynchronous=False, update_listener=None):
    port = port or config.PORT_CLOUDFORMATION
    log_startup_message('CloudFormation')
    backend_port = get_free_tcp_port()
    start_proxy_for_service('cloudformation', port, backend_port, update_listener)
    if RUN_SERVER_IN_PROCESS:
        cmd = '%s "%s" cloudformation -p %s -H 0.0.0.0' % (sys.executable, __file__, backend_port)
        env_vars = {'PYTHONPATH': ':'.join(sys.path)}
        return do_run(cmd, asynchronous, env_vars=env_vars)
    else:
        argv = ['cloudformation', '-p', str(backend_port), '-H', '0.0.0.0']
        thread = FuncThread(start_up, argv)
        thread.start()
        return thread


def set_moto_account_ids(resource_json):
    def fix_ids(obj, **kwargs):
        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, six.string_types):
                    if 'arn' in key.lower() or (':%s:' % TEST_AWS_ACCOUNT_ID) in value:
                        obj[key] = value.replace(TEST_AWS_ACCOUNT_ID, MOTO_ACCOUNT_ID)
        return obj

    return recurse_object(resource_json, fix_ids)


def get_entity_id(entity, resource_json=None):
    # check if physical_resource_id is present
    if hasattr(entity, 'physical_resource_id'):
        return entity.physical_resource_id
    # check ID attribute candidates
    types_with_ref_as_id_or_name = (apigw_models.RestAPI, apigw_models.Resource)
    attr_candidates = ['function_arn', 'Arn', 'Ref', 'id', 'Id', 'name', 'Name']
    for attr in attr_candidates:
        if hasattr(entity, attr):
            if attr in ['id', 'name'] and not isinstance(entity, types_with_ref_as_id_or_name):
                LOG.warning('Unable to find ARN, using "%s" instead: %s - %s', attr, resource_json, entity)
            return getattr(entity, attr)
        if hasattr(entity, 'get_cfn_attribute'):
            try:
                result = entity.get_cfn_attribute(attr)
                if result:
                    return result
            except Exception:
                pass
        if isinstance(entity, dict):
            if attr in entity:
                return entity.get(attr)
    # fall back to classes that use params as the dict of entity parameters
    if hasattr(entity, 'params'):
        for key, value in (entity.params or {}).items():
            if key.endswith('Name'):
                return value


def convert_objs_to_ids(resource_json):
    def fix_ids(obj, **kwargs):
        if isinstance(obj, dict):
            obj = dict(obj)
            for key, value in obj.items():
                if isinstance(value, BaseModel):
                    entity_id = get_entity_id(value)
                    obj[key] = entity_id or value
        return obj

    return recurse_object(resource_json, fix_ids)


def update_physical_resource_id(resource):
    phys_res_id = getattr(resource, 'physical_resource_id', None)
    if phys_res_id:
        return

    if isinstance(resource, lambda_models.LambdaFunction):
        func_arn = aws_stack.lambda_function_arn(resource.function_name)
        resource.function_arn = resource.physical_resource_id = func_arn

    elif isinstance(resource, service_models.StepFunctionsActivity):
        act_arn = aws_stack.stepfunctions_activity_arn(resource.params.get('Name'))
        resource.physical_resource_id = act_arn

    elif isinstance(resource, kinesis_models.Stream):
        resource.physical_resource_id = resource.stream_name

    elif isinstance(resource, service_models.LogsLogGroup):
        resource.physical_resource_id = resource.params.get('LogGroupName')

    elif isinstance(resource, service_models.FirehoseDeliveryStream):
        resource.physical_resource_id = resource.params.get('DeliveryStreamName')

    elif isinstance(resource, service_models.SecretsManagerSecret):
        resource.physical_resource_id = resource.params.get('Name')

    elif isinstance(resource, service_models.EventsRule):
        resource.physical_resource_id = resource.params.get('Name')

    elif isinstance(resource, service_models.ElasticsearchDomain):
        resource.physical_resource_id = resource.params.get('DomainName')

    elif isinstance(resource, dynamodb_models.Table):
        resource.physical_resource_id = resource.name

    elif isinstance(resource, dynamodb2_models.Table):
        resource.physical_resource_id = resource.name

    elif isinstance(resource, apigw_models.RestAPI):
        resource.physical_resource_id = resource.id

    elif isinstance(resource, apigw_models.Stage):
        resource.physical_resource_id = resource.get('stageName')

    elif isinstance(resource, apigw_models.Resource):
        resource.physical_resource_id = resource.id

    else:
        LOG.warning('Unable to determine physical_resource_id for resource %s' % type(resource))


def apply_attributes_from_existing_resource_on_update(resource_props, stack_name, existing_resource, resource_id):
    if not existing_resource or not resource_props:
        return

    res_type = resource_props['Type']
    props = resource_props.get('Properties', {})
    if res_type == 'AWS::S3::Bucket':
        existing_name = getattr(existing_resource, 'name', None)
        if existing_name:
            LOG.debug('Applying existing bucket name "%s" when updating resource %s' % (existing_name, resource_id))
            props['BucketName'] = existing_name

    if res_type == 'AWS::StepFunctions::StateMachine':
        existing_name = getattr(existing_resource, 'name', None)
        if existing_name:
            props['StateMachineName'] = existing_name


def apply_patches():
    """ Apply patches to make LocalStack seamlessly interact with the moto backend.
        TODO: Eventually, these patches should be contributed to the upstream repo! """

    # add model mappings to moto
    parsing.MODEL_MAP.update(MODEL_MAP)

    # fix account ID
    parsing.ACCOUNT_ID = TEST_AWS_ACCOUNT_ID

    # Patch clean_json in moto
    def clean_json(resource_json, resources_map):
        try:
            rs = clean_json_orig(resource_json, resources_map)
        except RecursionError:
            if isinstance(resource_json, dict) and 'Ref' in resource_json:
                LOG.info('Potential circular dependency detected when resolving Ref "%s"' % resource_json['Ref'])
                return resource_json['Ref']
            raise

        if isinstance(resource_json, dict):
            attr_ref = resource_json.get('Fn::GetAtt')
            if isinstance(attr_ref, list) and rs == resource_json:
                # If the attribute cannot be resolved (i.e., result == resource_json), and if indeed
                # the resource has not been deployed yet (entry in resources_map is None) then we raise
                # an exception here, which should cause the deployment loop to do another iteration
                map_keys = list(resources_map.keys())
                LOG.debug('Unable to resolve attribute reference %s in resource map keys %s' % (attr_ref, map_keys))
                if attr_ref[0] in map_keys and not resources_map[attr_ref[0]]:
                    raise DependencyNotYetSatisfied(
                        resource_ids=attr_ref[0],
                        message='Unable to resolve attribute ref %s' % attr_ref
                    )

                # If the attribute cannot be resolved for some other reason, then return
                # an empty value, to avoid returning the original JSON struct (which otherwise
                # results in downstream issues, e.g., when concatenating template values).
                return None

            if 'Ref' in resource_json and isinstance(rs, BaseModel):
                entity_id = get_entity_id(rs, resource_json)
                if entity_id:
                    return entity_id

                LOG.warning('Unable to resolve "Ref" attribute for: %s - %s - %s', resource_json, rs, type(rs))

            if 'Fn::Sub' in resource_json:
                if isinstance(resource_json['Fn::Sub'], list):
                    for k, v in resource_json['Fn::Sub'][1].items():
                        resource_json['Fn::Sub'][1][k] = clean_json(v, resources_map)

                    for key, val in resources_map._parsed_resources.items():
                        if not val:
                            continue

                        if not isinstance(val, str):
                            continue

                        for k, v in resource_json['Fn::Sub'][1].items():
                            if isinstance(v, dict) and 'Ref' in v:
                                if v['Ref'] == key:
                                    resource_json['Fn::Sub'][1][k] = val

                        resource_json['Fn::Sub'][0] = resource_json['Fn::Sub'][0].replace('${%s}' % key, val)

                    for k, v in resource_json['Fn::Sub'][1].items():
                        if v is not None:
                            resource_json['Fn::Sub'][0] = resource_json['Fn::Sub'][0].replace('${%s}' % k, str(v))

                    result = resource_json['Fn::Sub'][0]

                    stack_name = resources_map._parsed_resources['AWS::StackName']
                    result = template_deployer.resolve_placeholders_in_string(result,
                        stack_name=stack_name, resources=resources_map._resource_json_map)
                    return result

        return rs

    clean_json_orig = parsing.clean_json
    parsing.clean_json = clean_json

    def register_unresolved_refs(logical_id, resource_json, resources_map, required_resources, details=None):
        required_resources = required_resources or []
        existing = {k for k, v in resources_map._parsed_resources.items() if v is not None}
        LOG.info('Resource %s cannot be deployed, found unsatisfied dependencies. %s - %s (existing: %s)' % (
            logical_id, details, resource_json, existing))
        # details = [logical_id, resource_json, resources_map, region_name]
        unres_deps = resources_map._unresolved_resources = getattr(resources_map, '_unresolved_resources', {})
        unres_deps[logical_id] = unres_deps.get(logical_id) or []
        unres_deps[logical_id].extend(required_resources)

    # Patch parse_and_create_resource method in moto to deploy resources in LocalStack
    def parse_and_create_resource(logical_id, resource_json, resources_map, region_name, force_create=False):
        try:
            if hasattr(resources_map, '_deleted'):
                return

            result = _parse_and_create_resource(
                logical_id, resource_json, resources_map, region_name, force_create=force_create
            )
            return result
        except DependencyNotYetSatisfied as e:
            register_unresolved_refs(logical_id, resource_json, resources_map, e.resource_ids, details=e)
            raise
        except Exception as e:
            LOG.error('Unable to parse and create resource "%s": %s %s' % (logical_id, e, traceback.format_exc()))
            raise

    def parse_and_update_resource(logical_id, resource_json, resources_map, region_name):
        try:
            result = _parse_and_create_resource(logical_id, resource_json, resources_map, region_name, update=True)
            return result
        except DependencyNotYetSatisfied as e:
            register_unresolved_refs(logical_id, resource_json, resources_map, e.resource_ids, details=e)
            raise
        except Exception as e:
            LOG.error('Unable to parse and update resource "%s": %s %s' % (logical_id, e, traceback.format_exc()))
            raise

    def _parse_and_create_resource(
            logical_id, resource_json, resources_map, region_name, update=False, force_create=False
    ):
        stack_name = resources_map.get('AWS::StackName')
        resource_hash_key = (stack_name, logical_id)
        props = resource_json['Properties'] = resource_json.get('Properties') or {}

        # If the current stack is being updated, avoid infinite recursion
        updating = CURRENTLY_UPDATING_RESOURCES.get(resource_hash_key)
        LOG.debug('Currently processing (update=%s) resource %s/%s: %s' % (update, stack_name, logical_id, updating))
        if updating:
            return None

        # set updating flag for this resource
        CURRENTLY_UPDATING_RESOURCES[resource_hash_key] = True

        # check if this resource already exists in the resource map
        resource = resources_map._parsed_resources.get(logical_id)

        # apply existing attributes to resource details if they are already deployed
        if resource:
            apply_attributes_from_existing_resource_on_update(
                resource_json, stack_name, resource, resource_id=logical_id)

            apply_attributes_from_existing_resource_on_update(
                resources_map._resource_json_map.get(logical_id), stack_name, resource, resource_id=logical_id)

        if resource and not update and not force_create:
            return resource

        # check if all dependencies are satisfied
        resource_map_copy = dict(resources_map._resource_json_map)
        resource_map_copy.update(resources_map._template.get('Mappings', {}))
        unsatisfied_deps = template_deployer.get_unsatisfied_dependencies(
            logical_id, resource_map_copy, stack_name
        )
        if unsatisfied_deps:
            dep_keys = list(unsatisfied_deps.keys())
            register_unresolved_refs(logical_id, resource_json, resources_map, dep_keys, details=dep_keys)
            return None

        # parse and get final resource JSON
        resource_tuple = parsing.parse_resource_and_generate_name(logical_id, resource_json, resources_map)
        if not resource_tuple:
            return None

        _, resource_json, resource_name = resource_tuple

        # add some fixes and default props which otherwise cause deployments to fail
        add_default_resource_props(resource_json, stack_name, resource_name=resource_name, resource_id=logical_id)

        # fix resource ARNs, make sure to convert account IDs 000000000000 to 123456789012
        resource_json_arns_fixed = clone(json_safe(convert_objs_to_ids(resource_json)))
        set_moto_account_ids(resource_json_arns_fixed)
        template_deployer.remove_none_values(resource_json_arns_fixed)

        # create resource definition and store CloudFormation metadata in moto
        moto_create_error = None
        if (resource or update) and not force_create:
            _tmp = parse_and_update_resource_orig(logical_id, resource_json_arns_fixed, resources_map, region_name)
            resource = _tmp or resource

        elif not resource:
            try:
                resource = parse_and_create_resource_orig(
                    logical_id, resource_json_arns_fixed, resources_map, region_name
                )
                if not resource:
                    # this can happen if the resource has an associated Condition which evaluates to false
                    return resource
                resource.logical_id = logical_id
            except Exception as e:
                moto_create_error = e

        # get deployment state
        res_state = get_and_update_deployment_state(logical_id, resources_map._resource_json_map, stack_name, resource)

        # check whether this resource needs to be deployed
        resource_map_new = dict(resources_map._resource_json_map)
        resource_map_new.update(resources_map._template.get('Mappings', {}))
        resource_map_new[logical_id] = resource_json
        should_be_created = template_deployer.should_be_deployed(
            logical_id, resource_map_new, stack_name, deploy_state=res_state)

        # check for moto creation errors and raise an exception if needed
        if moto_create_error:
            if should_be_created:
                raise moto_create_error
            else:
                LOG.info('Error on moto CF resource creation for %s. Ignoring, as should_be_created=%s: %s' %
                         (logical_id, should_be_created, moto_create_error))

        # update physical_resource_id field
        if resource and not update:
            update_physical_resource_id(resource)

        # Fix for moto which sometimes hard-codes region name as 'us-east-1'
        if hasattr(resource, 'region_name') and resource.region_name != region_name:
            LOG.debug('Updating incorrect region from %s to %s' % (resource.region_name, region_name))
            resource.region_name = region_name

        # check whether this resource needs to be deployed
        is_updateable = False
        if not should_be_created:
            # This resource is either not deployable or already exists. Check if it can be updated
            is_updateable = template_deployer.is_updateable(logical_id, resource_map_new, stack_name)
            if not update or not is_updateable:
                unsatisfied_deps = template_deployer.get_unsatisfied_dependencies(
                    logical_id, resource_map_new, stack_name
                )
                if unsatisfied_deps:
                    register_unresolved_refs(logical_id, resource_json, resources_map, list(unsatisfied_deps.keys()))
                else:
                    LOG.debug('Resource %s need not be deployed (is_updateable=%s): %s %s' % (
                        logical_id, is_updateable, resource_json, bool(resource)))
                # Return if this resource already exists and can/need not be updated yet
                # NOTE: We should always return the resource here, to avoid duplicate
                #       creation of resources in moto!
                return resource

        # Deploy resource in LocalStack
        LOG.debug('Deploying CloudFormation resource (update=%s, exists=%s, updateable=%s): %s' %
                  (update, not should_be_created, is_updateable, resource_json))

        try:
            deploy_func = template_deployer.update_resource if update else template_deployer.deploy_resource
            result = deploy_func(logical_id, resource_map_new, stack_name=stack_name)
        finally:
            CURRENTLY_UPDATING_RESOURCES[resource_hash_key] = False

        if not should_be_created:
            # skip the parts below for update requests
            return resource

        def find_id(resource):
            """ Find ID of the given resource. """
            if not resource:
                return
            for id_attr in ('Id', 'id', 'ResourceId', 'RestApiId', 'DeploymentId', 'RoleId'):
                if id_attr in resource:
                    return resource[id_attr]

        # update resource IDs to avoid mismatch between CF moto and LocalStack backend resources
        if hasattr(resource, 'id') or (isinstance(resource, dict) and resource.get('id')):
            existing_id = resource.id if hasattr(resource, 'id') else resource['id']
            new_res_id = find_id(result)
            LOG.debug('Updating resource id: %s - %s, %s - %s' % (existing_id, new_res_id, resource, resource_json))
            if new_res_id:
                LOG.info('Updating resource ID from %s to %s (%s)' % (existing_id, new_res_id, region_name))
                update_resource_id(
                    resource, new_res_id, props, region_name, stack_name, resources_map._resource_json_map,
                    resource_json.get('Properties')
                )
            else:
                LOG.warning('Unable to extract id for resource %s: %s' % (logical_id, result))

        # update physical_resource_id field
        update_physical_resource_id(resource)

        return resource

    def get_and_update_deployment_state(resource_id, resources, stack_name, resource):
        """ Fetch and update the deployment state of the given stack resource. """
        res_details = resources[resource_id]
        details = template_deployer.get_deployment_state(resource_id, resources, stack_name)
        if details:
            if hasattr(resource, 'update_state'):
                resource.update_state(details)
            resource_props = res_details['Properties']
            if hasattr(resource, 'get_cfn_attribute') and not resource_props.get('PhysicalResourceId'):
                try:
                    resource_props['PhysicalResourceId'] = resource.get_cfn_attribute('Ref')
                except Exception:
                    # ignore this error here if the "Ref" attribute is not (yet) available
                    pass
        return details

    def update_resource_id(resource, new_id, props, region_name, stack_name, resource_map, resource_props={}):
        """ Update and fix the ID(s) of the given resource. """

        # NOTE: this is a bit of a hack, which is required because
        # of the order of events when CloudFormation resources are created.
        # When we process a request to create a CF resource that's part of a
        # stack, say, an API Gateway Resource, then we (1) create the object
        # in memory in moto, which generates a random ID for the resource, and
        # then (2) create the actual resource in the backend service using
        # template_deployer.deploy_resource(..) (see above).
        # The resource created in (2) now has a different ID than the resource
        # created in (1), which leads to downstream problems. Hence, we need
        # the logic below to reconcile the ids, i.e., apply IDs from (2) to (1).

        backend = apigw_models.apigateway_backends[region_name]
        if isinstance(resource, apigw_models.RestAPI):
            # We also need to fetch the resources to replace the root resource
            # that moto automatically adds to newly created RestAPI objects
            client = aws_stack.connect_to_service('apigateway')
            resources = client.get_resources(restApiId=new_id, limit=500)['items']
            # repoint ID mappings (make sure this stays BELOW calling get_resources() above)
            api = resource
            backend.apis.pop(api.id, None)
            api.id = new_id
            api.physical_resource_id = new_id
            backend.apis[new_id] = api
            # make sure no resources have been added in addition to the root /
            assert len(api.resources) == 1
            api.resources = {}
            for res in resources:
                res_path_part = res.get('pathPart') or res.get('path')
                res_method_path = resource_props.get('Body', {}).get('paths', {}).get(res_path_part)
                child = api.add_child(res_path_part, res.get('parentId'))

                for key in res_method_path or {}:
                    method_type = key.upper()
                    method = child.resource_methods.get(method_type)
                    if not method:
                        child.add_method(method_type, None, None)

                    path_int = res_method_path[key]['x-amazon-apigateway-integration']
                    child.add_integration(
                        method_type, path_int['type'], path_int.get('uri'),
                        request_templates=path_int.get('requestTemplates'),
                        integration_method=path_int.get('httpMethod'))

                api.resources.pop(child.id)
                child.id = res['id']
                child.api_id = new_id
                api.resources[child.id] = child

        elif isinstance(resource, apigw_models.Resource):
            api_id = props['RestApiId']
            api_id = template_deployer.resolve_refs_recursively(stack_name, api_id, resource_map)
            backend.apis[api_id].resources.pop(resource.id, None)
            backend.apis[api_id].resources[new_id] = resource
            resource.id = new_id
            if hasattr(resource, 'physical_resource_id'):
                resource.physical_resource_id = new_id

        elif isinstance(resource, apigw_models.Deployment):
            api_id = props['RestApiId']
            api_id = template_deployer.resolve_refs_recursively(stack_name, api_id, resource_map)
            if not api_id:
                api_id = resource_props['RestApiId']

            backend.apis[api_id].deployments.pop(resource['id'], None)
            backend.apis[api_id].deployments[new_id] = resource
            resource['id'] = new_id
            if hasattr(resource, 'physical_resource_id'):
                resource.physical_resource_id = new_id

        else:
            LOG.warning('Unexpected resource type when updating ID: %s' % type(resource))

    def parse_and_delete_resource(resource_name, resource_json, resources_map, region_name, *args, **kwargs):
        try:
            return parse_and_delete_resource_orig(resource_name, resource_json,
                resources_map, region_name, *args, **kwargs)
        except DependencyNotYetSatisfied:
            res_type = resource_json['Type']
            resource_class = parsing.resource_class_from_type(res_type)
            if resource_class:
                try:
                    resource_class.delete_from_cloudformation_json(resource_name, resource_json, region_name)
                except Exception as e:
                    LOG.info('Unable to delete resource "%s" (type %s): %s' % (resource_name, res_type, e))
        except AttributeError:
            # looks like a "delete" method is not yet implemented for a resource type -> ignore
            pass

    parse_and_create_resource_orig = parsing.parse_and_create_resource
    parsing.parse_and_create_resource = parse_and_create_resource
    parse_and_update_resource_orig = parsing.parse_and_update_resource
    parsing.parse_and_update_resource = parse_and_update_resource
    parse_and_delete_resource_orig = parsing.parse_and_delete_resource
    parsing.parse_and_delete_resource = parse_and_delete_resource

    def resource_map_delete(self, *args, **kwargs):
        self._deleted = True
        result = resource_map_delete_orig(self, *args, **kwargs)
        return result

    resource_map_delete_orig = parsing.ResourceMap.delete
    parsing.ResourceMap.delete = resource_map_delete

    def resource_map_create(self, template, *args, **kwargs):
        stack_name = self._parsed_resources['AWS::StackName']
        resources_json_map = template.get('Resources') or {}
        for res_id, res_details in resources_json_map.items():
            # add some fixes and default props which otherwise cause deployments to fail
            add_default_resource_props(res_details, stack_name, resource_id=res_id)
        result = resource_map_create_orig(self, template, *args, **kwargs)
        return result

    resource_map_create_orig = parsing.ResourceMap.create
    parsing.ResourceMap.create = resource_map_create

    def resource_map_update(self, template, *args, **kwargs):
        stack_name = self._parsed_resources['AWS::StackName']
        resources_json_map = template.get('Resources') or {}
        for res_id, res_details in resources_json_map.items():
            # add some fixes and default props which otherwise cause deployments to fail
            add_default_resource_props(res_details, stack_name, resource_id=res_id,
                existing_resources=self._resource_json_map)
        result = resource_map_update_orig(self, template, *args, **kwargs)
        return result

    resource_map_update_orig = parsing.ResourceMap.update
    parsing.ResourceMap.update = resource_map_update

    # patch ResourceMap set_resource_json()
    def set_resource_json(self, resources):
        resources = resources or {}
        for res_id, resource in resources.items():
            props = resource.get('Properties', {})
            for prop_name, prop_value in props.items():
                if isinstance(prop_value, dict):
                    if isinstance(prop_value.get('Fn::Sub'), str):
                        prop_value['Fn::Sub'] = [prop_value['Fn::Sub'], {}]

        self._resource_json_map = resources or {}
        self._resource_json_map_orig = clone_safe(self._resource_json_map)

    parsing.ResourceMap.set_resource_json = set_resource_json

    # patch CloudFormation parse_output(..) method to fix a bug in moto
    def parse_output(output_logical_id, output_json, resources_map):
        try:
            result = parse_output_orig(output_logical_id, output_json, resources_map)
        except KeyError:
            result = Output()
            result.key = output_logical_id
            result.value = None
            result.description = output_json.get('Description')

        # Make sure output includes export name
        if not hasattr(result, 'export_name'):
            export_name = output_json.get('Export', {}).get('Name')
            if export_name:
                result.export_name = clean_json(export_name, resources_map)

        return result

    parse_output_orig = parsing.parse_output
    parsing.parse_output = parse_output

    # Make sure the export name is returned for stack outputs
    if '<ExportName>' not in responses.DESCRIBE_STACKS_TEMPLATE:
        find = '</OutputValue>'
        replace = """</OutputValue>
        {% if output.export_name %}
        <ExportName>{{ output.export_name }}</ExportName>
        {% endif %}
        """
        responses.DESCRIBE_STACKS_TEMPLATE = responses.DESCRIBE_STACKS_TEMPLATE.replace(find, replace)

    resource_map_diff_orig = parsing.ResourceMap.diff

    def resource_map_diff(self, template, parameters=None):
        resources_diff = resource_map_diff_orig(self, template, parameters)
        if resources_diff['Add'] or resources_diff['Remove'] or resources_diff['Modify']:
            return resources_diff

        raise ValidationError(name_or_id='', message='No updates are to be performed.')

    parsing.ResourceMap.diff = resource_map_diff

    # Patch CloudFormationBackend.update_stack method in moto
    def make_cf_update_stack(cf_backend):
        cf_update_stack_orig = cf_backend.update_stack

        def cf_update_stack(self, *args, **kwargs):
            stack = cf_update_stack_orig(*args, **kwargs)
            # update stack exports
            self._validate_export_uniqueness(stack)
            for export in stack.exports:
                self.exports[export.name] = export

            return stack

        return types.MethodType(cf_update_stack, cf_backend)

    for region, cf_backend in cloudformation_backends.items():
        cf_backend.update_stack = make_cf_update_stack(cf_backend)

    # Patch DynamoDB get_cfn_attribute(..) method in moto
    def DynamoDB_Table_get_cfn_attribute(self, attribute_name):
        try:
            return ddb_table_get_cfn_attribute_orig(self, attribute_name)
        except Exception:
            if attribute_name == 'Arn':
                return aws_stack.dynamodb_table_arn(table_name=self.name)
            raise

    ddb_table_get_cfn_attribute_orig = dynamodb_models.Table.get_cfn_attribute
    dynamodb_models.Table.get_cfn_attribute = DynamoDB_Table_get_cfn_attribute

    # patch missing ":stream/" in Kinesis ARN

    @property
    def kinesis_arn(self, *args, **kwargs):
        result = self._arn_orig
        if ':stream/' not in result:
            parts = result.split(':')
            result = ':'.join(parts[:-1] + ['stream/%s' % parts[-1]])
        return result

    kinesis_models.Stream._arn_orig = kinesis_models.Stream.arn
    kinesis_models.Stream.arn = kinesis_arn

    # Patch generate_stack_id(..) method in moto
    def generate_stack_id(stack_name, region=None, **kwargs):
        region = region or aws_stack.get_region()
        return generate_stack_id_orig(stack_name, region=region, **kwargs)

    generate_stack_id_orig = cloudformation_utils.generate_stack_id
    cloudformation_utils.generate_stack_id = cloudformation_models.generate_stack_id = generate_stack_id

    # Patch DynamoDB get_cfn_attribute(..) method in moto
    def DynamoDB2_Table_get_cfn_attribute(self, attribute_name):
        if attribute_name == 'Arn':
            return aws_stack.dynamodb_table_arn(table_name=self.name)
        elif attribute_name == 'StreamArn':
            if (self.stream_specification or {}).get('StreamEnabled'):
                return aws_stack.dynamodb_stream_arn(self.name, 'latest')
            return None
        raise UnformattedGetAttTemplateException()

    dynamodb2_models.Table.get_cfn_attribute = DynamoDB2_Table_get_cfn_attribute

    # Patch SQS get_cfn_attribute(..) method in moto
    def SQS_Queue_get_cfn_attribute(self, attribute_name):
        if attribute_name in ['Arn', 'QueueArn']:
            return aws_stack.sqs_queue_arn(queue_name=self.name)
        return SQS_Queue_get_cfn_attribute_orig(self, attribute_name)

    SQS_Queue_get_cfn_attribute_orig = sqs_models.Queue.get_cfn_attribute
    sqs_models.Queue.get_cfn_attribute = SQS_Queue_get_cfn_attribute

    # Patch S3 Bucket get_cfn_attribute(..) method in moto
    def S3_Bucket_get_cfn_attribute(self, attribute_name):
        if attribute_name in ['Arn']:
            return aws_stack.s3_bucket_arn(self.name)
        return S3_Bucket_get_cfn_attribute_orig(self, attribute_name)

    S3_Bucket_get_cfn_attribute_orig = s3_models.FakeBucket.get_cfn_attribute
    s3_models.FakeBucket.get_cfn_attribute = S3_Bucket_get_cfn_attribute

    @classmethod
    def Bucket_update_from_cloudformation_json(cls,
            original_resource, new_resource_name, cloudformation_json, region_name):
        if cloudformation_json.get('BucketName') in [None, original_resource.name]:
            # TODO: apply other resource updates
            cloudformation_json['BucketName'] = original_resource.name
            return original_resource
        return Bucket_update_from_cf_json_orig(original_resource, new_resource_name, cloudformation_json, region_name)

    Bucket_update_from_cf_json_orig = s3_models.FakeBucket.update_from_cloudformation_json
    s3_models.FakeBucket.update_from_cloudformation_json = Bucket_update_from_cloudformation_json

    # Patch SQS physical_resource_id(..) method in moto
    @property
    def SQS_Queue_physical_resource_id(self):
        result = SQS_Queue_physical_resource_id_orig.fget(self)
        if '://' not in result:
            # convert ID to queue URL
            self._physical_resource_id = (getattr(self, '_physical_resource_id', None) or
                aws_stack.get_sqs_queue_url(result))
            return self._physical_resource_id
        return result

    SQS_Queue_physical_resource_id_orig = sqs_models.Queue.physical_resource_id
    sqs_models.Queue.physical_resource_id = SQS_Queue_physical_resource_id

    # Patch LogGroup get_cfn_attribute(..) method in moto
    def LogGroup_get_cfn_attribute(self, attribute_name):
        try:
            return LogGroup_get_cfn_attribute_orig(self, attribute_name)
        except Exception:
            if attribute_name == 'Arn':
                return aws_stack.log_group_arn(self.name)
            raise

    LogGroup_get_cfn_attribute_orig = getattr(cw_models.LogGroup, 'get_cfn_attribute', None)
    cw_models.LogGroup.get_cfn_attribute = LogGroup_get_cfn_attribute

    # Patch Lambda get_cfn_attribute(..) method in moto
    def Lambda_Function_get_cfn_attribute(self, attribute_name):
        try:
            if attribute_name == 'Arn':
                return self.function_arn
            return Lambda_Function_get_cfn_attribute_orig(self, attribute_name)
        except Exception:
            if attribute_name in ('Name', 'FunctionName'):
                return self.function_name
            raise

    Lambda_Function_get_cfn_attribute_orig = lambda_models.LambdaFunction.get_cfn_attribute
    lambda_models.LambdaFunction.get_cfn_attribute = Lambda_Function_get_cfn_attribute

    # Patch DynamoDB get_cfn_attribute(..) method in moto
    def DynamoDB_Table_get_cfn_attribute(self, attribute_name):
        try:
            if attribute_name == 'StreamArn':
                streams = aws_stack.connect_to_service('dynamodbstreams').list_streams(TableName=self.name)['Streams']
                return streams[0]['StreamArn'] if streams else None
            return DynamoDB_Table_get_cfn_attribute_orig(self, attribute_name)
        except Exception as e:
            LOG.warning('Unable to get attribute "%s" from resource %s: %s' % (attribute_name, type(self), e))
            raise

    DynamoDB_Table_get_cfn_attribute_orig = dynamodb_models.Table.get_cfn_attribute
    dynamodb_models.Table.get_cfn_attribute = DynamoDB_Table_get_cfn_attribute

    # Patch IAM get_cfn_attribute(..) method in moto
    def IAM_Role_get_cfn_attribute(self, attribute_name):
        try:
            return IAM_Role_get_cfn_attribute_orig(self, attribute_name)
        except Exception:
            if attribute_name == 'Arn':
                return aws_stack.role_arn(self.name)
            raise

    IAM_Role_get_cfn_attribute_orig = iam_models.Role.get_cfn_attribute
    iam_models.Role.get_cfn_attribute = IAM_Role_get_cfn_attribute

    # Patch IAM Role model
    # https://github.com/localstack/localstack/issues/925
    @property
    def IAM_Role_physical_resource_id(self):
        return self.name

    iam_models.Role.physical_resource_id = IAM_Role_physical_resource_id

    # Patch SNS Topic get_cfn_attribute(..) method in moto
    def SNS_Topic_get_cfn_attribute(self, attribute_name):
        result = SNS_Topic_get_cfn_attribute_orig(self, attribute_name)
        if attribute_name.lower() in ['arn', 'topicarn']:
            result = aws_stack.fix_account_id_in_arns(result)
        return result

    SNS_Topic_get_cfn_attribute_orig = sns_models.Topic.get_cfn_attribute
    sns_models.Topic.get_cfn_attribute = SNS_Topic_get_cfn_attribute

    # Patch create_from_cloudformation_json(..)
    # #2568 Cloudformation create-stack for SNS with yaml causes TypeError

    SNS_Topic_create_from_cloudformation_json_orig = sns_models.Topic.create_from_cloudformation_json

    def SNS_Topic_create_from_cloudformation_json(resource_name, cloudformation_json, region_name):
        properties = cloudformation_json['Properties']
        if properties.get('Subscription'):
            properties['Subscription'] = [subscription for subscription in properties['Subscription'] if subscription]
        result = SNS_Topic_create_from_cloudformation_json_orig(resource_name, cloudformation_json, region_name)
        # remove topic from backend, as it will be created by our mechanism with additional attributes (like tags)
        sns_models.sns_backends[region_name].topics.pop(result.arn)
        return result

    sns_models.Topic.create_from_cloudformation_json = SNS_Topic_create_from_cloudformation_json

    # Patch ES get_cfn_attribute(..) method
    def ES_get_cfn_attribute(self, attribute_name):
        if attribute_name in ['Arn', 'DomainArn']:
            return aws_stack.es_domain_arn(self.params.get('DomainName'))
        if attribute_name == 'DomainEndpoint':
            if not hasattr(self, '_domain_endpoint'):
                es_details = aws_stack.connect_to_service('es').describe_elasticsearch_domain(
                    DomainName=self.params.get('DomainName'))
                self._domain_endpoint = es_details['DomainStatus']['Endpoint']
            return self._domain_endpoint
        raise UnformattedGetAttTemplateException()

    service_models.ElasticsearchDomain.get_cfn_attribute = ES_get_cfn_attribute

    # Patch Firehose get_cfn_attribute(..) method
    def Firehose_get_cfn_attribute(self, attribute_name):
        if attribute_name == 'Arn':
            return aws_stack.firehose_stream_arn(self.params.get('DeliveryStreamName'))
        raise UnformattedGetAttTemplateException()

    service_models.FirehoseDeliveryStream.get_cfn_attribute = Firehose_get_cfn_attribute

    # Patch LambdaFunction create_from_cloudformation_json(..) method in moto
    @classmethod
    def Lambda_create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        resource_name = cloudformation_json.get('Properties', {}).get('FunctionName') or resource_name

        def _from_env(*args, **kwargs):
            result = Mock()
            result.api = Mock()
            result.api.get_adapter = (lambda *args, **kwargs: None)
            return result

        # Temporarily set a mock client, to prevent moto from talking to the Docker daemon
        import docker
        _from_env_orig = docker.from_env
        docker.from_env = _from_env

        try:
            result = Lambda_create_from_cloudformation_json_orig(resource_name, cloudformation_json, region_name)
        finally:
            docker.from_env = _from_env_orig
        return result

    Lambda_create_from_cloudformation_json_orig = lambda_models.LambdaFunction.create_from_cloudformation_json
    lambda_models.LambdaFunction.create_from_cloudformation_json = Lambda_create_from_cloudformation_json

    # Patch EventSourceMapping create_from_cloudformation_json(..) method in moto
    @classmethod
    def Mapping_create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        props = cloudformation_json.get('Properties', {})
        func_name = props.get('FunctionName') or ''
        if ':lambda:' in func_name:
            props['FunctionName'] = aws_stack.lambda_function_name(func_name)
        try:
            return Mapping_create_from_cloudformation_json_orig(resource_name, cloudformation_json, region_name)
        except Exception:
            LOG.info('Unable to add Lambda event mapping for source ARN "%s" in moto backend (ignoring)' %
                props.get('EventSourceArn'))
            # return an empty dummy instance, to avoid downstream None value issues
            return service_models.BaseModel()

    Mapping_create_from_cloudformation_json_orig = lambda_models.EventSourceMapping.create_from_cloudformation_json
    lambda_models.EventSourceMapping.create_from_cloudformation_json = Mapping_create_from_cloudformation_json

    # Patch LambdaFunction update_from_cloudformation_json(..) method in moto
    @classmethod
    def Lambda_update_from_cloudformation_json(cls,
            original_resource, new_resource_name, cloudformation_json, region_name):
        resource_name = cloudformation_json.get('Properties', {}).get('FunctionName') or new_resource_name
        return Lambda_create_from_cloudformation_json_orig(resource_name, cloudformation_json, region_name)

    if not hasattr(lambda_models.LambdaFunction, 'update_from_cloudformation_json'):
        lambda_models.LambdaFunction.update_from_cloudformation_json = Lambda_update_from_cloudformation_json

    # Patch Role update_from_cloudformation_json(..) method
    @classmethod
    def Role_update_from_cloudformation_json(cls,
            original_resource, new_resource_name, cloudformation_json, region_name):
        props = cloudformation_json.get('Properties', {})
        original_resource.name = props.get('RoleName') or original_resource.name
        original_resource.assume_role_policy_document = props.get('AssumeRolePolicyDocument')
        return original_resource

    if not hasattr(iam_models.Role, 'update_from_cloudformation_json'):
        iam_models.Role.update_from_cloudformation_json = Role_update_from_cloudformation_json

    # patch ApiGateway Deployment deletion
    @staticmethod
    def depl_delete_from_cloudformation_json(resource_name, resource_json, region_name):
        properties = resource_json['Properties']
        LOG.info('TODO: apigateway.Deployment.delete_from_cloudformation_json %s' % properties)

    if not hasattr(apigw_models.Deployment, 'delete_from_cloudformation_json'):
        apigw_models.Deployment.delete_from_cloudformation_json = depl_delete_from_cloudformation_json

    # patch/create ApiGateway Stage creation
    @staticmethod
    def stage_create_from_cloudformation_json(resource_name, resource_json, region_name):
        props = resource_json['Properties']
        name = props.get('StageName')
        description = props.get('Description') or ''
        deployment_id = props.get('DeploymentId')
        variables = props.get('Variables')
        return apigw_models.Stage(name=name, deployment_id=deployment_id, variables=variables, description=description)

    if not hasattr(apigw_models.Stage, 'create_from_cloudformation_json'):
        apigw_models.Stage.create_from_cloudformation_json = stage_create_from_cloudformation_json

    # patch Lambda Version deletion
    @staticmethod
    def vers_delete_from_cloudformation_json(resource_name, resource_json, region_name):
        properties = resource_json['Properties']
        LOG.info('TODO: apigateway.Deployment.delete_from_cloudformation_json %s' % properties)

    if not hasattr(lambda_models.LambdaVersion, 'delete_from_cloudformation_json'):
        lambda_models.LambdaVersion.delete_from_cloudformation_json = vers_delete_from_cloudformation_json

    # add CloudFormation types
    @classmethod
    def RestAPI_create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        props = cloudformation_json['Properties']
        name = props.get('Name') or short_uid()
        region_name = props.get('Region') or aws_stack.get_region()
        description = props.get('Description') or ''
        id = props.get('Id') or short_uid()
        return apigw_models.RestAPI(id, region_name, name, description)

    def RestAPI_get_cfn_attribute(self, attribute_name):
        if attribute_name == 'Id':
            return self.id
        if attribute_name == 'Region':
            return self.region_name
        if attribute_name == 'Name':
            return self.name
        if attribute_name == 'Description':
            return self.description
        if attribute_name == 'RootResourceId':
            for id, resource in self.resources.items():
                if resource.parent_id is None:
                    return resource.id
            return None
        raise UnformattedGetAttTemplateException()

    def Stage_get_cfn_attribute(self, attribute_name):
        if attribute_name == 'Ref':
            return self.get('stageName')
        raise super(apigw_models.Stage, self).get_cfn_attribute(attribute_name)

    @classmethod
    def Deployment_create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        props = cloudformation_json['Properties']
        deployment_id = short_uid()
        stage_name = props.get('StageName')  # fix upstream issue if StageName not present
        description = props.get('Description') or ''
        return apigw_models.Deployment(deployment_id, stage_name, description)

    @classmethod
    def Resource_create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        props = cloudformation_json['Properties']
        region_name = props.get('Region') or aws_stack.get_region()
        path_part = props.get('PathPart')
        api_id = props.get('RestApiId')
        parent_id = props.get('ParentId')
        id = props.get('Id') or short_uid()
        return apigw_models.Resource(id, region_name, api_id, path_part, parent_id)

    @classmethod
    def Method_create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        props = cloudformation_json['Properties']
        method_type = props.get('HttpMethod')
        authorization_type = props.get('AuthorizationType')
        return apigw_models.Method(method_type, authorization_type)

    def RestAPI_create_deployment(self, name, description='', stage_variables=None, **kwargs):
        name = name or '__tmp_stage__'
        result = RestAPI_create_deployment_orig(self, name, description='', stage_variables=None, **kwargs)
        self.stages.pop('__tmp_stage__', None)
        return result

    apigw_models.RestAPI.create_from_cloudformation_json = RestAPI_create_from_cloudformation_json
    apigw_models.RestAPI.get_cfn_attribute = RestAPI_get_cfn_attribute
    RestAPI_create_deployment_orig = apigw_models.RestAPI.create_deployment
    apigw_models.RestAPI.create_deployment = RestAPI_create_deployment
    apigw_models.Stage.get_cfn_attribute = Stage_get_cfn_attribute
    apigw_models.Deployment.create_from_cloudformation_json = Deployment_create_from_cloudformation_json
    apigw_models.Resource.create_from_cloudformation_json = Resource_create_from_cloudformation_json
    apigw_models.Method.create_from_cloudformation_json = Method_create_from_cloudformation_json
    # TODO: add support for AWS::ApiGateway::Model, AWS::ApiGateway::RequestValidator, ...

    # fix AttributeError in moto's CloudFormation describe_stack_resource
    def describe_stack_resource(self):
        stack_name = self._get_param('StackName')
        stack = self.cloudformation_backend.get_stack(stack_name)
        logical_resource_id = self._get_param('LogicalResourceId')
        if not stack:
            msg = ('Unable to find CloudFormation stack "%s" in region %s' %
                   (stack_name, aws_stack.get_region()))
            if aws_stack.get_region() != self.region:
                msg = '%s/%s' % (msg, self.region)
            LOG.warning(msg)
            response = aws_responses.flask_error_response_json(msg, code=404, error_type='ResourceNotFoundException')
            return 404, response.headers, response.data

        for stack_resource in stack.stack_resources:
            # Note: Line below has been patched
            # if stack_resource.logical_resource_id == logical_resource_id:
            if stack_resource and stack_resource.logical_resource_id == logical_resource_id:
                resource = stack_resource
                break
        else:
            raise ValidationError(stack_name,
                message='Unable to find resource "%s" in stack "%s"' % (logical_resource_id, stack_name))

        template = self.response_template(
            responses.DESCRIBE_STACK_RESOURCE_RESPONSE_TEMPLATE)
        return template.render(stack=stack, resource=resource)

    responses.CloudFormationResponse.describe_stack_resource = describe_stack_resource

    # fix moto's describe_stack_events jinja2.exceptions.UndefinedError
    def cf_describe_stack_events(self):
        stack_name = self._get_param('StackName')
        backend = self.cloudformation_backend
        stack = backend.get_stack(stack_name)
        if not stack:
            # Also return stack events for deleted stacks, specified by stack name
            stack = ([stk for id, stk in backend.deleted_stacks.items() if stk.name == stack_name] or [0])[0]
        if not stack:
            raise ValidationError(stack_name,
                message='Unable to find stack "%s" in region %s' % (stack_name, aws_stack.get_region()))
        template = self.response_template(responses.DESCRIBE_STACK_EVENTS_RESPONSE)
        return template.render(stack=stack)

    responses.CloudFormationResponse.describe_stack_events = cf_describe_stack_events

    # fix Lambda regions in moto - see https://github.com/localstack/localstack/issues/1961
    for region in boto3.session.Session().get_available_regions('lambda'):
        if region not in lambda_models.lambda_backends:
            lambda_models.lambda_backends[region] = lambda_models.LambdaBackend(region)

    def initialize_resources(self):
        self.resource_map._template = self.resource_map._template or self.template_dict
        try:
            self.resource_map.load()
            self.resource_map.create(self.resource_map._template)
        except Exception:
            # looks like a dependency issue in the first iteration of creating resource_map - should
            # be fixed by run_dependencies_deployment_loop(..) below
            pass
        run_dependencies_deployment_loop(self, 'CREATE', initialize=True)

    def stack_update(self, template, role_arn=None, parameters=None, tags=None, **kwargs):
        # Note: code is copied in here, we want to prevent calling the original method (as it fires state events)
        self._add_stack_event('UPDATE_IN_PROGRESS', resource_status_reason='User Initiated')
        self.template = template
        self._parse_template()
        self.resource_map.update(self.template_dict, parameters)
        self.output_map = self._create_output_map()
        self.role_arn = role_arn
        if tags is not None:
            self.tags = tags
        # now run our deployment loop
        run_dependencies_deployment_loop(self, 'UPDATE')

    @property
    def stack_outputs(self):
        if self.status.endswith('_IN_PROGRESS'):
            # avoid iterating the resource_map if stack creation is still in progress
            return []
        return stack_outputs_orig.__get__(self)

    FakeStack.initialize_resources = initialize_resources
    FakeStack.update = stack_update
    stack_outputs_orig = FakeStack.stack_outputs
    FakeStack.stack_outputs = stack_outputs

    # patch FakeChangeSet constructor to pass parent stack parameters and resource map

    def change_set_init(self, stack_id, stack_name, stack_template, change_set_id,
            change_set_name, template, parameters, region_name, *args, **kwargs):
        stack = [s for s in cloudformation_backends[region_name].stacks.values() if s.name == stack_name]
        if stack:
            stack = stack[0]
            stack_params = stack.parameters or {}
            stack_params.update(parameters or {})
            parameters = stack_params

        # call parent constructor
        change_set_init_orig(self, stack_id, stack_name, stack_template, change_set_id,
                change_set_name, template, parameters, region_name, *args, **kwargs)

        # update resource map with resources from parent stack
        if stack:
            self.resource_map._parsed_resources.update(stack.resource_map._parsed_resources)

    change_set_init_orig = FakeChangeSet.__init__
    FakeChangeSet.__init__ = change_set_init

    # patch Kinesis Stream get_cfn_attribute(..) method in moto
    def Kinesis_Stream_get_cfn_attribute(self, attribute_name):
        if attribute_name == 'Arn':
            return self.arn

        raise UnformattedGetAttTemplateException()

    kinesis_models.Stream.get_cfn_attribute = Kinesis_Stream_get_cfn_attribute

    # patch cloudformation backend create_change_set(..)
    # #760 cloudformation deploy invalid xml error
    cloudformation_backend_create_change_set_orig = CloudFormationBackend.create_change_set

    def cloudformation_backend_create_change_set(
            self,
            stack_name,
            change_set_name,
            template,
            parameters,
            region_name,
            change_set_type,
            notification_arns=None,
            tags=None,
            role_arn=None):
        change_set_id, _ = cloudformation_backend_create_change_set_orig(
            self,
            stack_name,
            change_set_name,
            template,
            parameters,
            region_name,
            change_set_type,
            notification_arns,
            tags,
            role_arn
        )
        change_set = self.change_sets[change_set_id]
        change_set.status = 'CREATE_COMPLETE'

        return change_set_id, _

    CloudFormationBackend.create_change_set = cloudformation_backend_create_change_set

    # patch _validate_export_uniqueness to avoid traversing resources before they are deployed

    def make_validate_export_uniqueness(cf_backend):
        def _validate_export_uniqueness(self, stack, *args, **kwargs):
            if not stack.output_map._output_json_map:
                return
            new_exports = [val.get('Export', {}).get('Name')
                for val in stack.output_map._output_json_map.values() if val.get('Export')]
            for export in self.exports.values():
                if export.name in new_exports and export.exporting_stack_id != stack.stack_id:
                    LOG.info('Conflicting export name "%s" (stack "%s") - already exported from stack "%s"' % (
                        export.name, stack.name, export.exporting_stack_id))
                    raise ValidationError(stack.stack_id,
                        message='Export names must be unique across a given region')
        return types.MethodType(_validate_export_uniqueness, cf_backend)

    for region, cf_backend in cloudformation_backends.items():
        cf_backend._validate_export_uniqueness = make_validate_export_uniqueness(cf_backend)

    # patch cloudformation backend set_exports
    # #3375 CloudFormation - All deployed stacks have StackStatus: CREATE_FAILED
    # Actually this patch has provided in moto-ext 1.3.15.45
    def make_set_exports(backend):
        def cf_backend_set_exports(self, new_stack):
            for export in new_stack.exports:
                self.exports[export.name] = export

        return types.MethodType(cf_backend_set_exports, backend)

    if not hasattr(CloudFormationBackend, 'set_exports'):
        for region, cf_backend in cloudformation_backends.items():
            cf_backend.set_exports = make_set_exports(cf_backend)

    # patch cloudformation backend change_set methods
    # #2240 - S3 bucket not created since 0.10.8
    # #2568 - Cloudformation create-stack for SNS with yaml causes TypeError
    cloudformation_backend_describe_change_set_orig = CloudFormationBackend.describe_change_set
    cloudformation_backend_execute_change_set_orig = CloudFormationBackend.execute_change_set

    def cloudformation_backend_describe_change_set(self, change_set_name, stack_name=None):
        change_set_name = change_set_name.replace(TEST_AWS_ACCOUNT_ID, MOTO_CFN_ACCOUNT_ID)
        return cloudformation_backend_describe_change_set_orig(self, change_set_name, stack_name)

    def cloudformation_backend_execute_change_set(self, change_set_name, stack_name=None):
        change_set_name = change_set_name.replace(TEST_AWS_ACCOUNT_ID, MOTO_CFN_ACCOUNT_ID)

        stack = self.change_sets.get(change_set_name)
        if not stack:
            for cs in self.change_sets:
                if self.change_sets[cs].change_set_name == change_set_name:
                    stack = self.change_sets[cs]

        def do_execute(*args):
            try:
                cloudformation_backend_execute_change_set_orig(self, change_set_name, stack_name)
                stack.output_map = stack._create_output_map()
                for export in stack.exports:
                    self.exports[export.name] = export
                set_stack_status(stack, 'CREATE_COMPLETE')
            except Exception:
                set_stack_status(stack, 'CREATE_FAILED')
                raise

        # start execution in background thread, to avoid timeouts/retries from the client
        set_stack_status(stack, 'CREATE_IN_PROGRESS')
        start_thread(do_execute)
        return True

    CloudFormationBackend.describe_change_set = cloudformation_backend_describe_change_set
    CloudFormationBackend.execute_change_set = cloudformation_backend_execute_change_set


def set_stack_status(stack, status):
    stack._add_stack_event(status)
    stack.status = status


# patch FakeStack.initialize_resources
def run_dependencies_deployment_loop(stack, action, initialize=False):
    def run_loop(*args):
        result = {}
        try:
            loop_result = do_run_loop()
            if isinstance(loop_result, parsing.ResourceMap):
                if initialize:
                    # create output map
                    stack.output_map.create()
                    # set exported resources
                    cloudformation_backends[stack.region_name].set_exports(stack)
                return
            result = loop_result
        except Exception as e:
            LOG.info('Error running stack deployment loop: %s' % e)
        set_stack_status(stack, '%s_FAILED' % action)
        raise Exception('Unable to resolve all CloudFormation resources after traversing ' +
            'dependency tree (maximum depth %s): %s' % (MAX_DEPENDENCY_DEPTH, list(result.keys())))

    def do_run_loop():
        # NOTE: We're adding this additional loop, as it seems that in some cases moto
        #   does not consider resource dependencies (e.g., if a "DependsOn" resource property
        #   is defined). This loop allows us to incrementally resolve such dependencies.
        resource_map = stack.resource_map
        unresolved = {}
        for i in range(MAX_DEPENDENCY_DEPTH):
            CURRENTLY_UPDATING_RESOURCES.clear()
            LOG.debug('Running CloudFormation stack deployment loop iteration %s' % (i + 1))
            try:
                # try iterating all resources in the map, catch any errors that may happen
                resource_map.values()
            except Exception as e:
                LOG.debug('Temporary error in loop iteration %s: %s' % (i + 1, e))
            unresolved = getattr(resource_map, '_unresolved_resources', {})
            LOG.debug('Found %s unresolved dependencies in loop iteration %s' % (len(unresolved), i + 1))
            if not unresolved:
                set_stack_status(stack, '%s_COMPLETE' % action)
                return resource_map
            resource_map._unresolved_resources = {}
            for resource_id, dependencies in unresolved.items():
                # Re-trigger the resource creation: simply access the key in the
                # map, which will cause the resource creation internally
                for dep_resource_id in dependencies:
                    resource_map[dep_resource_id]
                resource_map[resource_id]
            if unresolved.keys() == resource_map._unresolved_resources.keys():
                # looks like no more resources can be resolved -> bail
                break
        LOG.info('Unresolvable dependencies, there may be undeployed stack resources: %s' % unresolved)
        return unresolved

    # NOTE: We're running the loop in the background, as it might take some time to complete
    FuncThread(run_loop).start()


def inject_stats_endpoint():
    """ Inject a simple /_stats endpoint into the moto server backend Web app. """
    # TODO: move this utility method to a shared file and enable it for all API endpoints
    from moto import server as moto_server

    def _get_stats():
        from pympler import muppy, summary
        all_objects = muppy.get_objects()
        result = summary.summarize(all_objects)
        result = result[0:20]
        summary = '\n'.join([line for line in summary.format_(result)])
        result = '%s\n\n%s' % (summary, json.dumps(result))
        return result, 200, {'content-type': 'text/plain'}

    def create_backend_app(service):
        backend_app = moto_server.create_backend_app_orig(service)
        backend_app.add_url_rule(
            '/_stats', endpoint='_get_stats', methods=['GET'], view_func=_get_stats, strict_slashes=False)
        return backend_app

    if not hasattr(moto_server, 'create_backend_app_orig'):
        moto_server.create_backend_app_orig = moto_server.create_backend_app
        moto_server.create_backend_app = create_backend_app


def start_up(*args):
    # patch moto implementation
    apply_patches()

    # add memory profiling endpoint
    inject_stats_endpoint()

    return moto_main(*args)


def main():
    setup_logging()

    # make sure all API names and ports are mapped properly
    canonicalize_api_names()

    # start API
    sys.exit(start_up())


if __name__ == '__main__':
    main()



import os
import json
import glob
import tempfile
import time
import requests
import shutil
import zipfile
import importlib
from six import iteritems
from localstack.utils.aws import aws_stack
from localstack.constants import (
    LOCALSTACK_ROOT_FOLDER, LOCALSTACK_VENV_FOLDER, LAMBDA_TEST_ROLE, TEST_AWS_ACCOUNT_ID, ENV_INTERNAL_TEST_RUN)
from localstack.utils.common import TMP_FILES, run, mkdir, to_str, load_file, save_file, is_alpine
from localstack.services.awslambda.lambda_utils import (
    get_handler_file_from_name, LAMBDA_DEFAULT_HANDLER, LAMBDA_DEFAULT_RUNTIME, LAMBDA_DEFAULT_STARTING_POSITION)

ARCHIVE_DIR_PREFIX = 'lambda.archive.'
DEFAULT_GET_LOG_EVENTS_DELAY = 3
LAMBDA_TIMEOUT_SEC = 6


def is_local_test_mode():
    """ Whether we are running in the context of our local integration tests. """
    return bool(os.environ.get(ENV_INTERNAL_TEST_RUN))


def copy_dir(source, target):
    if is_alpine():
        # Using the native command can be an order of magnitude faster on Travis-CI
        return run('cp -r %s %s' % (source, target))
    shutil.copytree(source, target)


def rm_dir(dir):
    if is_alpine():
        # Using the native command can be an order of magnitude faster on Travis-CI
        return run('rm -r %s' % dir)

    shutil.rmtree(dir)


def create_lambda_archive(script, get_content=False, libs=[], runtime=None, file_name=None):
    """ Utility method to create a Lambda function archive """
    runtime = runtime or LAMBDA_DEFAULT_RUNTIME
    tmp_dir = tempfile.mkdtemp(prefix=ARCHIVE_DIR_PREFIX)
    TMP_FILES.append(tmp_dir)
    file_name = file_name or get_handler_file_from_name(LAMBDA_DEFAULT_HANDLER, runtime=runtime)
    script_file = os.path.join(tmp_dir, file_name)
    if os.path.sep in script_file:
        mkdir(os.path.dirname(script_file))
        # create __init__.py files along the path to allow Python imports
        path = file_name.split(os.path.sep)
        for i in range(1, len(path)):
            save_file(os.path.join(tmp_dir, *(path[:i] + ['__init__.py'])), '')
    save_file(script_file, script)
    # copy libs
    for lib in libs:
        paths = [lib, '%s.py' % lib]
        try:
            module = importlib.import_module(lib)
            paths.append(module.__file__)
        except Exception:
            pass
        target_dir = tmp_dir
        root_folder = os.path.join(LOCALSTACK_VENV_FOLDER, 'lib/python*/site-packages')
        if lib == 'localstack':
            paths = ['localstack/*.py', 'localstack/utils']
            root_folder = LOCALSTACK_ROOT_FOLDER
            target_dir = os.path.join(tmp_dir, lib)
            mkdir(target_dir)
        for path in paths:
            file_to_copy = path if path.startswith('/') else os.path.join(root_folder, path)
            for file_path in glob.glob(file_to_copy):
                name = os.path.join(target_dir, file_path.split(os.path.sep)[-1])
                if os.path.isdir(file_path):
                    copy_dir(file_path, name)
                else:
                    shutil.copyfile(file_path, name)

    # create zip file
    result = create_zip_file(tmp_dir, get_content=get_content)
    return result


def delete_lambda_function(name):
    client = aws_stack.connect_to_service('lambda')
    client.delete_function(FunctionName=name)


def create_zip_file_cli(source_path, base_dir, zip_file):
    # Using the native zip command can be an order of magnitude faster on Travis-CI
    source = '*' if source_path == base_dir else os.path.basename(source_path)
    command = 'cd %s; zip -r %s %s' % (base_dir, zip_file, source)
    run(command)


def create_zip_file_python(source_path, base_dir, zip_file):
    with zipfile.ZipFile(zip_file, 'w') as zip_file:
        for root, dirs, files in os.walk(base_dir):
            for name in files:
                full_name = os.path.join(root, name)
                relative = root[len(base_dir):].lstrip(os.path.sep)
                dest = os.path.join(relative, name)
                zip_file.write(full_name, dest)


def create_zip_file(file_path, zip_file=None, get_content=False):
    base_dir = file_path
    if not os.path.isdir(file_path):
        base_dir = tempfile.mkdtemp(prefix=ARCHIVE_DIR_PREFIX)
        shutil.copy(file_path, base_dir)
        TMP_FILES.append(base_dir)
    tmp_dir = tempfile.mkdtemp(prefix=ARCHIVE_DIR_PREFIX)
    full_zip_file = zip_file
    if not full_zip_file:
        zip_file_name = 'archive.zip'
        full_zip_file = os.path.join(tmp_dir, zip_file_name)
    # create zip file
    if is_alpine():
        create_zip_file_cli(file_path, base_dir, zip_file=full_zip_file)
    else:
        create_zip_file_python(file_path, base_dir, zip_file=full_zip_file)
    if not get_content:
        TMP_FILES.append(tmp_dir)
        return full_zip_file
    zip_file_content = None
    with open(full_zip_file, 'rb') as file_obj:
        zip_file_content = file_obj.read()
    rm_dir(tmp_dir)
    return zip_file_content


def create_lambda_function(func_name, zip_file=None, event_source_arn=None, handler_file=None,
        handler=None, starting_position=None, runtime=None, envvars={},
        tags={}, libs=[], delete=False, layers=None, **kwargs):
    """Utility method to create a new function via the Lambda API"""

    starting_position = starting_position or LAMBDA_DEFAULT_STARTING_POSITION
    runtime = runtime or LAMBDA_DEFAULT_RUNTIME
    client = aws_stack.connect_to_service('lambda')

    # load zip file content if handler_file is specified
    if not zip_file and handler_file:
        if libs or not handler:
            zip_file = create_lambda_archive(load_file(handler_file), libs=libs,
                get_content=True, runtime=runtime or LAMBDA_DEFAULT_RUNTIME)
        else:
            zip_file = create_zip_file(handler_file, get_content=True)

    handler = handler or LAMBDA_DEFAULT_HANDLER

    if delete:
        try:
            # Delete function if one already exists
            client.delete_function(FunctionName=func_name)
        except Exception:
            pass

    # create function
    additional_kwargs = kwargs
    kwargs = {
        'FunctionName': func_name,
        'Runtime': runtime,
        'Handler': handler,
        'Role': LAMBDA_TEST_ROLE,
        'Code': {
            'ZipFile': zip_file
        },
        'Timeout': LAMBDA_TIMEOUT_SEC,
        'Environment': dict(Variables=envvars),
        'Tags': tags
    }
    kwargs.update(additional_kwargs)
    if layers:
        kwargs['Layers'] = layers
    create_func_resp = client.create_function(**kwargs)

    resp = {
        'CreateFunctionResponse': create_func_resp,
        'CreateEventSourceMappingResponse': None
    }

    # create event source mapping
    if event_source_arn:
        resp['CreateEventSourceMappingResponse'] = client.create_event_source_mapping(
            FunctionName=func_name,
            EventSourceArn=event_source_arn,
            StartingPosition=starting_position
        )

    return resp


def connect_api_gateway_to_http_with_lambda_proxy(gateway_name, target_uri,
        stage_name=None, methods=[], path=None, auth_type=None, http_method=None):
    if not methods:
        methods = ['GET', 'POST', 'DELETE']
    if not path:
        path = '/'
    stage_name = stage_name or 'test'
    resources = {}
    resource_path = path.lstrip('/')
    resources[resource_path] = []
    for method in methods:
        int_meth = http_method or method
        resources[resource_path].append({
            'httpMethod': method,
            'authorizationType': auth_type,
            'integrations': [{
                'type': 'AWS_PROXY',
                'uri': target_uri,
                'httpMethod': int_meth
            }]
        })
    return aws_stack.create_api_gateway(
        name=gateway_name,
        resources=resources,
        stage_name=stage_name
    )


def assert_objects(asserts, all_objects):
    if type(asserts) is not list:
        asserts = [asserts]
    for obj in asserts:
        assert_object(obj, all_objects)


def assert_object(expected_object, all_objects):
    # for Python 3 compatibility
    dict_values = type({}.values())
    if isinstance(all_objects, dict_values):
        all_objects = list(all_objects)
    # wrap single item in an array
    if type(all_objects) is not list:
        all_objects = [all_objects]
    found = find_object(expected_object, all_objects)
    if not found:
        raise Exception('Expected object not found: %s in list %s' % (expected_object, all_objects))


def find_object(expected_object, object_list):
    for obj in object_list:
        if isinstance(obj, list):
            found = find_object(expected_object, obj)
            if found:
                return found

        all_ok = True
        if obj != expected_object:
            if not isinstance(expected_object, dict):
                all_ok = False
            else:
                for k, v in iteritems(expected_object):
                    if not find_recursive(k, v, obj):
                        all_ok = False
                        break
        if all_ok:
            return obj
    return None


def find_recursive(key, value, obj):
    if isinstance(obj, dict):
        for k, v in iteritems(obj):
            if k == key and v == value:
                return True
            if find_recursive(key, value, v):
                return True
    elif isinstance(obj, list):
        for o in obj:
            if find_recursive(key, value, o):
                return True
    else:
        return False


def list_all_s3_objects():
    return map_all_s3_objects().values()


def delete_all_s3_objects(buckets):
    s3_client = aws_stack.connect_to_service('s3')
    buckets = buckets if isinstance(buckets, list) else [buckets]
    for bucket in buckets:
        keys = all_s3_object_keys(bucket)
        deletes = [{'Key': key} for key in keys]
        if deletes:
            s3_client.delete_objects(Bucket=bucket, Delete={'Objects': deletes})


def download_s3_object(s3, bucket, path):
    with tempfile.SpooledTemporaryFile() as tmpfile:
        s3.Bucket(bucket).download_fileobj(path, tmpfile)
        tmpfile.seek(0)
        result = tmpfile.read()
        try:
            result = to_str(result)
        except Exception:
            pass
        return result


def all_s3_object_keys(bucket):
    s3_client = aws_stack.connect_to_resource('s3')
    bucket = s3_client.Bucket(bucket) if isinstance(bucket, str) else bucket
    keys = [key for key in bucket.objects.all()]
    return keys


def map_all_s3_objects(to_json=True, buckets=None):
    s3_client = aws_stack.connect_to_resource('s3')
    result = {}
    buckets = buckets if not buckets or isinstance(buckets, list) else [buckets]
    buckets = [s3_client.Bucket(b) for b in buckets] if buckets else s3_client.buckets.all()
    for bucket in buckets:
        for key in bucket.objects.all():
            value = download_s3_object(s3_client, key.bucket_name, key.key)
            try:
                if to_json:
                    value = json.loads(value)
                key = '%s%s%s' % (key.bucket_name, '' if key.key.startswith('/') else '/', key.key)
                result[key] = value
            except Exception:
                # skip non-JSON or binary objects
                pass
    return result


def get_sample_arn(service, resource):
    return 'arn:aws:%s:%s:%s:%s' % (service, aws_stack.get_region(), TEST_AWS_ACCOUNT_ID, resource)


def send_describe_dynamodb_ttl_request(table_name):
    return send_dynamodb_request('', 'DescribeTimeToLive', json.dumps({'TableName': table_name}))


def send_update_dynamodb_ttl_request(table_name, ttl_status):
    return send_dynamodb_request('', 'UpdateTimeToLive', json.dumps({
        'TableName': table_name,
        'TimeToLiveSpecification': {
            'AttributeName': 'ExpireItem',
            'Enabled': ttl_status
        }
    }))


def send_dynamodb_request(path, action, request_body):
    headers = {
        'Host': 'dynamodb.amazonaws.com',
        'x-amz-target': 'DynamoDB_20120810.{}'.format(action),
        'authorization': 'some_token'
    }
    url = '{}/{}'.format(os.getenv('TEST_DYNAMODB_URL'), path)
    return requests.put(url, data=request_body, headers=headers, verify=False)


def create_sqs_queue(queue_name):
    """Utility method to create a new queue via SQS API"""

    client = aws_stack.connect_to_service('sqs')

    # create queue
    queue_url = client.create_queue(QueueName=queue_name)['QueueUrl']

    # get the queue arn
    queue_arn = client.get_queue_attributes(
        QueueUrl=queue_url,
        AttributeNames=['QueueArn'],
    )['Attributes']['QueueArn']

    return {
        'QueueUrl': queue_url,
        'QueueArn': queue_arn,
    }


def get_lambda_log_group_name(function_name):
    return '/aws/lambda/{}'.format(function_name)


def check_expected_lambda_log_events_length(expected_length, function_name):
    events = get_lambda_log_events(function_name)
    events = [line for line in events if line not in ['\x1b[0m', '\\x1b[0m']]
    if len(events) != expected_length:
        print('Invalid # of Lambda %s log events: %s / %s: %s' % (function_name, len(events), expected_length, events))
    assert len(events) == expected_length
    return events


def get_lambda_log_events(function_name, delay_time=DEFAULT_GET_LOG_EVENTS_DELAY):
    def get_log_events(function_name, delay_time):
        time.sleep(delay_time)

        logs = aws_stack.connect_to_service('logs')
        log_group_name = get_lambda_log_group_name(function_name)
        rs = logs.filter_log_events(logGroupName=log_group_name)

        return rs['events']

    try:
        events = get_log_events(function_name, delay_time)
    except Exception as e:
        if 'ResourceNotFoundException' in str(e):
            return []
        raise

    rs = []
    for event in events:
        raw_message = event['message']
        if not raw_message or 'START' in raw_message or 'END' in raw_message or 'REPORT' in raw_message:
            continue
        if raw_message in ['\x1b[0m', '\\x1b[0m']:
            continue

        try:
            rs.append(json.loads(raw_message))
        except Exception:
            rs.append(raw_message)

    return rs



from moto.cloudwatch.models import cloudwatch_backends
from localstack.utils.aws import aws_stack, aws_responses
from localstack.utils.common import replace_response_content, parse_request_data
from localstack.utils.tagging import TaggingService
from localstack.services.generic_proxy import ProxyListener

XMLNS_CLOUDWATCH = ''

# path for backdoor API to receive raw metrics
PATH_GET_RAW_METRICS = '/cloudwatch/metrics/raw'

TAGS = TaggingService()


class ProxyListenerCloudWatch(ProxyListener):

    def forward_request(self, method, path, data, headers):
        req_data = parse_request_data(method, path, data)
        action = req_data.get('Action')
        if action == 'TagResource':
            arn = req_data.get('ResourceARN')
            tags = aws_stack.extract_tags(req_data)
            TAGS.tag_resource(arn, tags)
            return aws_responses.requests_response_xml(action, {}, xmlns=XMLNS_CLOUDWATCH)
        if action == 'UntagResource':
            arn = req_data.get('ResourceARN')
            tag_names = [v for k, v in req_data.items() if k.startswith('TagKeys.member.')]
            TAGS.untag_resource(arn, tag_names)
            return aws_responses.requests_response_xml(action, {}, xmlns=XMLNS_CLOUDWATCH)
        if action == 'ListTagsForResource':
            arn = req_data.get('ResourceARN')
            tags = TAGS.list_tags_for_resource(arn)
            result = {'Tags': {'member': tags.get('Tags', [])}}
            return aws_responses.requests_response_xml(action, result, xmlns=XMLNS_CLOUDWATCH)
        if path.startswith(PATH_GET_RAW_METRICS):
            result = cloudwatch_backends[aws_stack.get_region()].metric_data
            result = [
                {'ns': r.namespace, 'n': r.name, 'v': r.value, 't': r.timestamp,
                 'd': [{'n': d.name, 'v': d.value} for d in r.dimensions]}
                for r in result
            ]
            return {'metrics': result}
        return True

    def return_response(self, method, path, data, headers, response):

        req_data = parse_request_data(method, path, data)
        action = req_data.get('Action')
        if action == 'PutMetricAlarm':
            name = req_data.get('AlarmName')
            # add missing attribute "TreatMissingData"
            treat_missing_data = req_data.get('TreatMissingData', 'ignore')
            cloudwatch_backends[aws_stack.get_region()].alarms[name].treat_missing_data = treat_missing_data
            # record tags
            arn = aws_stack.cloudwatch_alarm_arn(name)
            tags = aws_stack.extract_tags(req_data)
            if tags:
                TAGS.tag_resource(arn, tags)

        # Fix Incorrect date format to the correct format
        # the dictionary contains the tag as the key and the value is a
        # tuple (pattern, replacement)

        regexes1 = (r'<{}>([^<]+) ([^<+]+)(\+[^<]*)?</{}>', r'<{}>\1T\2Z</{}>')
        regexes2 = (r'<{}>([^<]+) ([^<+.]+)(\.[^<]*)?</{}>', r'<{}>\1T\2Z</{}>')
        timestamp_tags = {
            'AlarmConfigurationUpdatedTimestamp': regexes1,
            'StateUpdatedTimestamp': regexes1,
            'member': regexes2
        }

        for tag, value in timestamp_tags.items():
            pattern, replacement = value
            self.fix_date_format(response, tag, pattern, replacement)
        response.headers['Content-Length'] = len(response.content)
        return response

    def fix_date_format(self, response, timestamp_tag, pattern, replacement):
        """ Normalize date to correct format"""
        pattern = pattern.format(timestamp_tag, timestamp_tag)
        replacement = replacement.format(timestamp_tag, timestamp_tag)
        replace_response_content(response, pattern, replacement)


# instantiate listener
UPDATE_CLOUD_WATCH = ProxyListenerCloudWatch()



import json
import logging
import traceback
import xmltodict
from flask import Flask, request
from requests.models import Response
from localstack.utils.aws import aws_stack
from localstack.utils.common import (
    parse_request_data, short_uid, long_uid, clone, clone_safe, select_attributes,
    timestamp_millis, recurse_object)
from localstack.utils.cloudformation import template_deployer, template_preparer
from localstack.services.generic_proxy import RegionBackend
from localstack.utils.aws.aws_responses import (
    requests_response_xml, requests_to_flask_response, flask_error_response_xml)

APP_NAME = 'cloudformation_api'
app = Flask(APP_NAME)

LOG = logging.getLogger(__name__)

XMLNS_CF = 'http://cloudformation.amazonaws.com/doc/2010-05-15/'


class CloudFormationRegion(RegionBackend):

    def __init__(self):
        # maps stack ID to stack details
        self.stacks = {}

    @property
    def exports(self):
        exports = []
        output_keys = {}
        for stack_id, stack in self.stacks.items():
            for output in stack.outputs:
                export_name = output.get('ExportName')
                if not export_name:
                    continue
                if export_name in output_keys:
                    # TODO: raise exception on stack creation in case of duplicate exports
                    LOG.warning('Found duplicate export name %s in stacks: %s %s' % (
                        export_name, output_keys[export_name], stack.stack_id))
                entry = {
                    'ExportingStackId': stack.stack_id,
                    'Name': export_name,
                    'Value': output['OutputValue']
                }
                exports.append(entry)
                output_keys[export_name] = stack.stack_id
        return exports


class Stack(object):
    def __init__(self, metadata=None, template={}):
        self.metadata = metadata or {}
        self.template = template or {}
        self._template_raw = clone_safe(self.template)
        self.template_original = clone_safe(self.template)
        # initialize resources
        for resource_id, resource in self.template_resources.items():
            resource['LogicalResourceId'] = self.template_original['Resources'][resource_id]['LogicalResourceId'] = (
                resource.get('LogicalResourceId') or resource_id)
        # initialize stack template attributes
        self.template['StackId'] = self.metadata['StackId'] = (self.metadata.get('StackId') or
            aws_stack.cloudformation_stack_arn(self.stack_name, short_uid()))
        self.template['Parameters'] = self.template.get('Parameters') or {}
        self.template['Outputs'] = self.template.get('Outputs') or {}
        # initialize metadata
        self.metadata['Parameters'] = self.metadata.get('Parameters') or []
        self.metadata['StackStatus'] = 'CREATE_IN_PROGRESS'
        self.metadata['CreationTime'] = self.metadata.get('CreationTime') or timestamp_millis()
        # maps resource id to resource state
        self.resource_states = {}
        # maps resource id to moto resource class instance (TODO: remove in the future)
        self.moto_resource_statuses = {}
        # list of stack events
        self.events = []
        # list of stack change sets
        self.change_sets = []
        # initialize parameters
        for i in range(1, 100):
            key = 'Parameters.member.%s.ParameterKey' % i
            value = 'Parameters.member.%s.ParameterValue' % i
            key = self.metadata.get(key)
            value = self.metadata.get(value)
            if not key:
                break
            self.metadata['Parameters'].append({'ParameterKey': key, 'ParameterValue': value})

    def describe_details(self):
        attrs = ['StackId', 'StackName', 'Description', 'StackStatusReason',
            'StackStatus', 'Capabilities', 'ParentId', 'RootId', 'RoleARN',
            'CreationTime', 'DeletionTime', 'LastUpdatedTime', 'ChangeSetId']
        result = select_attributes(self.metadata, attrs)
        result['Tags'] = self.tags
        result['Outputs'] = self.outputs
        result['Parameters'] = self.stack_parameters()
        for attr in ['Capabilities', 'Tags', 'Outputs', 'Parameters']:
            result[attr] = {'member': result.get(attr, [])}
        return result

    def set_stack_status(self, status):
        self.metadata['StackStatus'] = status
        self.add_stack_event(self.stack_name, self.stack_id, status)

    def add_stack_event(self, resource_id, physical_res_id, status):
        event = {
            'EventId': long_uid(),
            'Timestamp': timestamp_millis(),
            'StackId': self.stack_id,
            'StackName': self.stack_name,
            'LogicalResourceId': resource_id,
            'PhysicalResourceId': physical_res_id,
            'ResourceStatus': status,
            'ResourceType': 'AWS::CloudFormation::Stack'
        }
        self.events.insert(0, event)

    def set_resource_status(self, resource_id, status, physical_res_id=None):
        resource = self.resources[resource_id]
        state = self.resource_states[resource_id] = self.resource_states.get(resource_id) or {}
        attr_defaults = (('LogicalResourceId', resource_id), ('PhysicalResourceId', physical_res_id))
        for res in [resource, state]:
            for attr, default in attr_defaults:
                res[attr] = res.get(attr) or default
        state['ResourceStatus'] = status
        state['StackName'] = state.get('StackName') or self.stack_name
        state['StackId'] = state.get('StackId') or self.stack_id
        state['ResourceType'] = state.get('ResourceType') or self.resources[resource_id].get('Type')
        self.add_stack_event(resource_id, physical_res_id, status)

    def resource_status(self, resource_id):
        result = self._lookup(self.resource_states, resource_id)
        return result

    @property
    def stack_name(self):
        return self.metadata['StackName']

    @property
    def stack_id(self):
        return self.metadata['StackId']

    @property
    def resources(self):
        """ Return dict of resources, parameters, conditions, and other stack metadata. """

        def add_params(defaults=True):
            for param in self.stack_parameters(defaults=defaults):
                if param['ParameterKey'] not in result:
                    props = {'Value': param['ParameterValue']}
                    result[param['ParameterKey']] = {
                        'Type': 'Parameter',
                        'LogicalResourceId': param['ParameterKey'],
                        'Properties': props
                    }

        result = dict(self.template_resources)

        add_params(defaults=False)

        for name, value in self.conditions.items():
            if name not in result:
                result[name] = {'Type': 'Parameter', 'LogicalResourceId': name, 'Properties': {'Value': value}}
        for name, value in self.mappings.items():
            if name not in result:
                result[name] = {'Type': 'Parameter', 'LogicalResourceId': name, 'Properties': {'Value': value}}

        add_params(defaults=True)

        return result

    @property
    def template_resources(self):
        return self.template['Resources']

    @property
    def tags(self):
        return aws_stack.extract_tags(self.metadata)

    @property
    def imports(self):
        def _collect(o, **kwargs):
            if isinstance(o, dict):
                import_val = o.get('Fn::ImportValue')
                if import_val:
                    result.add(import_val)
            return o
        result = set()
        recurse_object(self.resources, _collect)
        return result

    @property
    def outputs(self):
        result = []
        # first, fetch the outputs of nested child stacks
        for stack in self.nested_stacks:
            result.extend(stack.outputs)
        # now, fetch the outputs of this stack
        for k, details in self.template.get('Outputs', {}).items():
            value = None
            try:
                template_deployer.resolve_refs_recursively(self.stack_name, details, self.resources)
                value = details['Value']
            except Exception as e:
                LOG.debug('Unable to resolve references in stack outputs: %s - %s' % (details, e))
            exports = details.get('Export') or {}
            export = exports.get('Name')
            description = details.get('Description')
            entry = {'OutputKey': k, 'OutputValue': value, 'Description': description, 'ExportName': export}
            result.append(entry)
        return result

    def stack_parameters(self, defaults=True):
        result = {}
        # add default template parameter values
        if defaults:
            for key, value in self.template_parameters.items():
                result[key] = {'ParameterKey': key, 'ParameterValue': value.get('Default')}
        # add stack parameters
        result.update({p['ParameterKey']: p for p in self.metadata['Parameters']})
        # add parameters of change sets
        for change_set in self.change_sets:
            result.update({p['ParameterKey']: p for p in change_set.metadata['Parameters']})
        result = list(result.values())
        return result

    @property
    def template_parameters(self):
        return self.template['Parameters']

    @property
    def conditions(self):
        return self.template.get('Conditions', {})

    @property
    def mappings(self):
        return self.template.get('Mappings', {})

    @property
    def exports_map(self):
        result = {}
        for export in CloudFormationRegion.get().exports:
            result[export['Name']] = export
        return result

    @property
    def nested_stacks(self):
        """ Return a list of nested stacks that have been deployed by this stack. """
        result = [r for r in self.template_resources.values() if r['Type'] == 'AWS::CloudFormation::Stack']
        result = [find_stack(r['Properties'].get('StackName')) for r in result]
        result = [r for r in result if r]
        return result

    @property
    def status(self):
        return self.metadata['StackStatus']

    @property
    def resource_types(self):
        return [r.get('Type') for r in self.template_resources.values()]

    def resource(self, resource_id):
        return self._lookup(self.resources, resource_id)

    def _lookup(self, resource_map, resource_id):
        resource = resource_map.get(resource_id)
        if not resource:
            raise Exception('Unable to find details for resource "%s" in stack "%s"' % (resource_id, self.stack_name))
        return resource

    def copy(self):
        return Stack(metadata=dict(self.metadata), template=dict(self.template))


class StackChangeSet(Stack):
    def __init__(self, params={}, template={}):
        super(StackChangeSet, self).__init__(params, template)
        name = self.metadata['ChangeSetName']
        if not self.metadata.get('ChangeSetId'):
            self.metadata['ChangeSetId'] = aws_stack.cf_change_set_arn(name, change_set_id=short_uid())

        stack = self.stack = find_stack(self.metadata['StackName'])
        self.metadata['StackId'] = stack.stack_id
        self.metadata['Status'] = 'CREATE_PENDING'

    @property
    def change_set_id(self):
        return self.metadata['ChangeSetId']

    @property
    def change_set_name(self):
        return self.metadata['ChangeSetName']

    @property
    def resources(self):
        result = dict(self.stack.resources)
        result.update(self.resources)
        return result

    @property
    def changes(self):
        result = self.metadata['Changes'] = self.metadata.get('Changes', [])
        return result


# --------------
# API ENDPOINTS
# --------------

def create_stack(req_params):
    state = CloudFormationRegion.get()
    template_deployer.prepare_template_body(req_params)
    template = template_preparer.parse_template(req_params['TemplateBody'])
    stack_name = template['StackName'] = req_params.get('StackName')
    stack = Stack(req_params, template)

    # find existing stack with same name, and remove it if this stack is in DELETED state
    existing = ([s for s in state.stacks.values() if s.stack_name == stack_name] or [None])[0]
    if existing:
        if 'DELETE' not in existing.status:
            return error_response('Stack named "%s" already exists with status "%s"' % (
                stack_name, existing.status), code=400, code_string='ValidationError')
        state.stacks.pop(existing.stack_id)

    state.stacks[stack.stack_id] = stack
    LOG.debug('Creating stack "%s" with %s resources ...' % (stack.stack_name, len(stack.template_resources)))
    deployer = template_deployer.TemplateDeployer(stack)
    try:
        deployer.deploy_stack()
    except Exception as e:
        stack.set_stack_status('CREATE_FAILED')
        msg = 'Unable to create stack "%s": %s' % (stack.stack_name, e)
        LOG.debug('%s %s' % (msg, traceback.format_exc()))
        return error_response(msg, code=400, code_string='ValidationError')
    result = {'StackId': stack.stack_id}
    return result


def delete_stack(req_params):
    stack_name = req_params.get('StackName')
    stack = find_stack(stack_name)
    deployer = template_deployer.TemplateDeployer(stack)
    deployer.delete_stack()
    return {}


def update_stack(req_params):
    stack_name = req_params.get('StackName')
    stack = find_stack(stack_name)
    if not stack:
        return error_response('Unable to update non-existing stack "%s"' % stack_name,
            code=404, code_string='ValidationError')
    template_preparer.prepare_template_body(req_params)
    template = template_preparer.parse_template(req_params['TemplateBody'])
    new_stack = Stack(req_params, template)
    deployer = template_deployer.TemplateDeployer(stack)
    try:
        deployer.update_stack(new_stack)
    except Exception as e:
        stack.set_stack_status('UPDATE_FAILED')
        msg = 'Unable to update stack "%s": %s' % (stack_name, e)
        LOG.debug('%s %s' % (msg, traceback.format_exc()))
        return error_response(msg, code=400, code_string='ValidationError')
    result = {'StackId': stack.stack_id}
    return result


def describe_stacks(req_params):
    state = CloudFormationRegion.get()
    stack_name = req_params.get('StackName')
    stacks = [s.describe_details() for s in state.stacks.values() if stack_name in [None, s.stack_name, s.stack_id]]
    if stack_name and not stacks:
        return error_response('Stack with id %s does not exist' % stack_name,
            code=400, code_string='ValidationError')
    result = {'Stacks': {'member': stacks}}
    return result


def list_stacks(req_params):
    state = CloudFormationRegion.get()
    filter = req_params.get('StackStatusFilter')
    stacks = [s.describe_details() for s in state.stacks.values() if filter in [None, s.status]]
    attrs = ['StackId', 'StackName', 'TemplateDescription', 'CreationTime', 'LastUpdatedTime', 'DeletionTime',
        'StackStatus', 'StackStatusReason', 'ParentId', 'RootId', 'DriftInformation']
    stacks = [select_attributes(stack, attrs) for stack in stacks]
    result = {'StackSummaries': {'member': stacks}}
    return result


def describe_stack_resource(req_params):
    stack_name = req_params.get('StackName')
    resource_id = req_params.get('LogicalResourceId')
    stack = find_stack(stack_name)
    if not stack:
        return stack_not_found_error(stack_name)
    details = stack.resource_status(resource_id)
    result = {'StackResourceDetail': details}
    return result


def describe_stack_resources(req_params):
    stack_name = req_params.get('StackName')
    resource_id = req_params.get('LogicalResourceId')
    phys_resource_id = req_params.get('PhysicalResourceId')
    if phys_resource_id and stack_name:
        return error_response('Cannot specify both StackName and PhysicalResourceId')
    # TODO: filter stack by PhysicalResourceId!
    stack = find_stack(stack_name)
    if not stack:
        return stack_not_found_error(stack_name)
    statuses = [stack.resource_status(res_id) for res_id, _ in stack.resource_states.items() if
        resource_id in [res_id, None]]
    return {'StackResources': {'member': statuses}}


def list_stack_resources(req_params):
    result = describe_stack_resources(req_params)
    if not isinstance(result, dict):
        return result
    result = {'StackResourceSummaries': result.pop('StackResources')}
    return result


def create_change_set(req_params):
    stack_name = req_params.get('StackName')
    template_deployer.prepare_template_body(req_params)
    template = template_preparer.parse_template(req_params.pop('TemplateBody'))
    template['StackName'] = stack_name
    template['ChangeSetName'] = req_params.get('ChangeSetName')
    stack = existing = find_stack(stack_name)
    if not existing:
        # automatically create (empty) stack if none exists yet
        state = CloudFormationRegion.get()
        empty_stack_template = dict(template)
        empty_stack_template['Resources'] = {}
        req_params_copy = clone_stack_params(req_params)
        stack = Stack(req_params_copy, empty_stack_template)
        state.stacks[stack.stack_id] = stack
        stack.set_stack_status('CREATE_COMPLETE')
    change_set = StackChangeSet(req_params, template)
    deployer = template_deployer.TemplateDeployer(stack)
    deployer.construct_changes(stack, change_set, change_set_id=change_set.change_set_id, append_to_changeset=True)
    stack.change_sets.append(change_set)
    change_set.metadata['Status'] = 'CREATE_COMPLETE'
    return {'StackId': change_set.stack_id, 'Id': change_set.change_set_id}


def execute_change_set(req_params):
    stack_name = req_params.get('StackName')
    cs_name = req_params.get('ChangeSetName')
    change_set = find_change_set(cs_name, stack_name=stack_name)
    if not change_set:
        return error_response('Unable to find change set "%s" for stack "%s"' % (cs_name, stack_name))
    LOG.debug('Executing change set "%s" for stack "%s" with %s resources ...' % (
        cs_name, stack_name, len(change_set.template_resources)))
    deployer = template_deployer.TemplateDeployer(change_set.stack)
    deployer.apply_change_set(change_set)
    change_set.stack.metadata['ChangeSetId'] = change_set.change_set_id
    return {}


def list_change_sets(req_params):
    stack_name = req_params.get('StackName')
    stack = find_stack(stack_name)
    if not stack:
        return error_response('Unable to find stack "%s"' % stack_name)
    result = [cs.metadata for cs in stack.change_sets]
    result = {'Summaries': {'member': result}}
    return result


def describe_change_set(req_params):
    stack_name = req_params.get('StackName')
    cs_name = req_params.get('ChangeSetName')
    change_set = find_change_set(cs_name, stack_name=stack_name)
    if not change_set:
        return error_response('Unable to find change set "%s" for stack "%s"' % (cs_name, stack_name))
    return change_set.metadata


def list_exports(req_params):
    state = CloudFormationRegion.get()
    result = {'Exports': {'member': state.exports}}
    return result


def list_imports(req_params):
    state = CloudFormationRegion.get()
    export_name = req_params.get('ExportName')
    importing_stack_names = []
    for stack in state.stacks.values():
        if export_name in stack.imports:
            importing_stack_names.append(stack.stack_name)
    result = {'Imports': {'member': importing_stack_names}}
    return result


def validate_template(req_params):
    try:
        result = template_preparer.validate_template(req_params)
        result = '<tmp>%s</tmp>' % result
        result = xmltodict.parse(result)['tmp']
        return result
    except Exception as err:
        return error_response('Template Validation Error: %s' % err)


def describe_stack_events(req_params):
    stack_name = req_params.get('StackName')
    state = CloudFormationRegion.get()
    events = []
    for stack_id, stack in state.stacks.items():
        if stack_name in [None, stack.stack_name, stack.stack_id]:
            events.extend(stack.events)
    return {'StackEvents': {'member': events}}


def delete_change_set(req_params):
    stack_name = req_params.get('StackName')
    cs_name = req_params.get('ChangeSetName')
    change_set = find_change_set(cs_name, stack_name=stack_name)
    if not change_set:
        return error_response('Unable to find change set "%s" for stack "%s"' % (cs_name, stack_name))
    change_set.stack.change_sets = [cs for cs in change_set.stack.change_sets if cs.change_set_name != cs_name]
    return {}


def get_template(req_params):
    stack_name = req_params.get('StackName')
    cs_name = req_params.get('ChangeSetName')
    stack = find_stack(stack_name)
    if cs_name:
        stack = find_change_set(stack_name, cs_name)
    if not stack:
        return stack_not_found_error(stack_name)
    result = {'TemplateBody': json.dumps(stack._template_raw)}
    return result


def get_template_summary(req_params):
    stack_name = req_params.get('StackName')
    stack = None
    if stack_name:
        stack = find_stack(stack_name)
        if not stack:
            return stack_not_found_error(stack_name)
    else:
        template_deployer.prepare_template_body(req_params)
        template = template_preparer.parse_template(req_params['TemplateBody'])
        req_params['StackName'] = 'tmp-stack'
        stack = Stack(req_params, template)
    result = stack.describe_details()
    id_summaries = {}
    for resource_id, resource in stack.template_resources.items():
        res_type = resource['Type']
        id_summaries[res_type] = id_summaries.get(res_type) or []
        id_summaries[res_type].append(resource_id)
    result['ResourceTypes'] = list(id_summaries.keys())
    result['ResourceIdentifierSummaries'] = [
        {'ResourceType': key, 'LogicalResourceIds': {'member': values}} for key, values in id_summaries.items()]
    return result


# -----------------
# MAIN ENTRY POINT
# -----------------

@app.route('/', methods=['POST'])
def handle_request():
    data = request.get_data()
    req_params = parse_request_data(request.method, request.path, data)
    action = req_params.get('Action', '')

    func = ENDPOINTS.get(action)
    if not func:
        return '', 404
    result = func(req_params)

    result = _response(action, result)
    return result


ENDPOINTS = {
    'CreateChangeSet': create_change_set,
    'CreateStack': create_stack,
    'DeleteChangeSet': delete_change_set,
    'DeleteStack': delete_stack,
    'DescribeChangeSet': describe_change_set,
    'DescribeStackEvents': describe_stack_events,
    'DescribeStackResource': describe_stack_resource,
    'DescribeStackResources': describe_stack_resources,
    'DescribeStacks': describe_stacks,
    'ExecuteChangeSet': execute_change_set,
    'GetTemplate': get_template,
    'GetTemplateSummary': get_template_summary,
    'ListChangeSets': list_change_sets,
    'ListExports': list_exports,
    'ListImports': list_imports,
    'ListStacks': list_stacks,
    'ListStackResources': list_stack_resources,
    'UpdateStack': update_stack,
    'ValidateTemplate': validate_template
}


# ---------------
# UTIL FUNCTIONS
# ---------------

def error_response(*args, **kwargs):
    kwargs['xmlns'] = kwargs.get('xmlns') or XMLNS_CF
    return flask_error_response_xml(*args, **kwargs)


def stack_not_found_error(stack_name):
    return error_response('Unable to find stack named "%s"' % stack_name,
        code=404, code_string='ResourceNotFoundException')


def clone_stack_params(stack_params):
    try:
        return clone(stack_params)
    except Exception as e:
        LOG.info('Unable to clone stack parameters: %s' % e)
        return stack_params


def find_stack(stack_name):
    state = CloudFormationRegion.get()
    return ([s for s in state.stacks.values() if stack_name in [s.stack_name, s.stack_id]] or [None])[0]


def find_change_set(cs_name, stack_name=None):
    state = CloudFormationRegion.get()
    stack = find_stack(stack_name)
    stacks = [stack] if stack else state.stacks.values()
    result = [cs for s in stacks for cs in s.change_sets if cs_name in [cs.change_set_id, cs.change_set_name]]
    return (result or [None])[0]


def _response(action, result):
    if isinstance(result, (dict, str)):
        result = requests_response_xml(action, result, xmlns=XMLNS_CF)
    if isinstance(result, Response):
        result = requests_to_flask_response(result)
    return result


def serve(port, quiet=True):
    from localstack.services import generic_proxy  # moved here to fix circular import errors
    return generic_proxy.serve_flask_app(app=app, port=port, quiet=quiet)
